/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Scanner;
import java.util.TreeSet;

import edu.umn.cs.kumarbio.Chromosome;
import edu.umn.cs.kumarbio.Genome;
import edu.umn.cs.kumarbio.shear.heterogeneity.Breakpoint;
import edu.umn.cs.kumarbio.shear.heterogeneity.SvHeterogeneityEstimator;
import edu.umn.cs.kumarbio.shear.realignment.BreakpointsRealigner;
import edu.umn.cs.kumarbio.shear.system.RunGatkCommand;
import edu.umn.cs.kumarbio.shear.system.RunPicardCommand;
import edu.umn.cs.kumarbio.shear.system.RunSystemCommand;
import edu.umn.cs.kumarbio.shear.variantcallers.CrestWrapper;
import edu.umn.cs.kumarbio.shear.variantcallers.DellyWrapper;
import edu.umn.cs.kumarbio.shear.variantcallers.HaplotypeCallerWrapper;
import edu.umn.cs.kumarbio.shear.variantcallers.VariantCaller;
import edu.umn.cs.kumarbio.shear.variants.Deletion;
import edu.umn.cs.kumarbio.shear.variants.Insertion;
import edu.umn.cs.kumarbio.shear.variants.Inversion;
import edu.umn.cs.kumarbio.shear.variants.SNP;
import edu.umn.cs.kumarbio.shear.variants.TandemDuplication;
import edu.umn.cs.kumarbio.shear.variants.Translocation;
import edu.umn.cs.kumarbio.shear.variants.Variant;

/**
 * This is used to run SHEAR-SV. The constructor creates a new ShearSV instance
 * and parses all of the provided command-line arguments. The program is run
 * using the execute method.
 */
public class ShearSV {

  // Prefix for all generated files
  private String  prefix;

  // Path to BAM alignment file to use
  private String  origBamFile;

  // Path to FASTA reference file to use
  private String  origFaRefFile;

  // SV prediction algorithm to use
  private String  svAlgo;

  // Alignment algorithm to use for local realignment
  private String  alignAlgo;

  // Path to *.2bit reference file to use (if SV algorithm is CREST)
  private String  origTwoBitRefFile;

  // Prefix of the BWA index files to use
  private String  origBwaIndexPrefix;

  // Region of the input alignment to analyze
  private String  region;

  // Path to pre-existing SV prediction file to use
  private String  origPreds;

  // Minimum hetergeneity level threshhold
  private double  minHetThresh;

  // Whether or not to preprocess input BAM file
  private boolean preprocess;

  // Whether or not SV-Only mode is activated
  private boolean svOnly;

  // Whether or not to realign unmapped reads during local realignment
  private boolean realignUnmapped;

  // Whether or not debugging mode is activated
  private boolean debug;

  // Whether or not Heterogeneity-Only mode is activated (hidden feature)
  private boolean hetOnly;

  // Path to PICARD directory
  private String  picardDir;

  // Path to CREST directory
  private String  crestDir;

  // Read group information
  private String  rgId;
  private String  rgLb;
  private String  rgPl;
  private String  rgPu;
  private String  rgSm;

  // Average length of the reads in the BAM alignment file to use
  private int     readLength;

  // Genome object to use
  private Genome  gen;

  /**
   * Creates a new ShearSV object by initializing the local variables that are
   * read from the command line arguments.
   *
   * @param args
   *          The command line arguments to process, excluding the first one
   *          (SHEAR command, i.e. 'sv' or 'assemble')
   */
  public ShearSV(String[] args) throws ShearUsageException {

    // Initialize some option values
    String minHetThreshString = null;
    minHetThresh = 0d;
    preprocess = false;
    svOnly = false;
    realignUnmapped = false;
    debug = false;
    hetOnly = false;

    // Read in command-line arguments
    for (int i = 0; i < args.length; i++) {
      try {
        if (args[i].equals("-p")) {
          prefix = args[++i];
        } else if (args[i].equals("-b")) {
          origBamFile = args[++i];
        } else if (args[i].equals("-f")) {
          origFaRefFile = args[++i];
        } else if (args[i].equals("--sv-algo")) {
          svAlgo = args[++i];
        } else if (args[i].equals("--align-algo")) {
          alignAlgo = args[++i];
        } else if (args[i].equals("-t")) {
          origTwoBitRefFile = args[++i];
        } else if (args[i].equals("-i")) {
          origBwaIndexPrefix = args[++i];
        } else if (args[i].equals("-r")) {
          region = args[++i];
        } else if (args[i].equals("--preds")) {
          origPreds = args[++i];
        } else if (args[i].equals("--min-het")) {
          minHetThreshString = args[++i];
        } else if (args[i].equals("--preprocess")) {
          preprocess = true;
        } else if (args[i].equals("--sv-only")) {
          svOnly = true;
        } else if (args[i].equals("--realign-unmapped")) {
          realignUnmapped = true;
        } else if (args[i].equals("-d")) {
          debug = true;
        } else if (args[i].equals("--het-only")) {
          hetOnly = true;
        } else {
          throw new ShearUsageException("Unknown argument: " + args[i]);
        }
      } catch (ArrayIndexOutOfBoundsException e) {
        throw new ShearUsageException("Missing argument value for flag '" + args[i - 1] + "'");
      }
    }

    // Read in ENV variables
    picardDir = System.getenv("PICARD_DIRECTORY");
    crestDir = System.getenv("CREST_DIRECTORY");

    try {

      // Validate prefix
      if (prefix == null) {
        throw new ShearUsageException("Must provide a prefix with the '-p' argument");
      }
      if (!prefix.matches("^[a-zA-Z0-9_\\-\\.]+$")) {
        throw new ShearUsageException(
            "Prefix must consist of alphanumeric characters, underscores, dashes, or periods.");
      }

      // Validate origBamFile
      if (origBamFile == null) {
        throw new ShearUsageException("Must provide a BAM alignment file with the '-b' argument");
      }
      if (!new File(origBamFile).exists()) {
        throw new ShearUsageException("BAM alignment file does not exist: " + origBamFile);
      } else if (!new File(origBamFile + ".bai").exists()) {
        throw new ShearUsageException("BAM alignment index file does not exist: " + origBamFile + ".bai");
      } else {
        origBamFile = new File(origBamFile).getCanonicalPath();
      }

      // Validate origFaRefFile
      if (origFaRefFile == null) {
        throw new ShearUsageException("Must provide a FASTA reference file with the '-f' argument");
      }
      if (!new File(origFaRefFile).exists()) {
        throw new ShearUsageException("FASTA reference file does not exist: " + origFaRefFile);
      } else if (!new File(origFaRefFile + ".fai").exists()) {
        throw new ShearUsageException("FASTA reference index file does not exist: " + origFaRefFile + ".fai");
      } else {
        origFaRefFile = new File(origFaRefFile).getCanonicalPath();
      }

      // Validate svAlgo
      if (svAlgo == null) {
        svAlgo = "delly";
      }
      if (!svAlgo.matches("^crest|delly$")) {
        throw new ShearUsageException("Invalid SV prediction algorithm specified: " + svAlgo);
      }

      // Validate alignAlgo
      if (alignAlgo == null) {
        alignAlgo = "bwamem";
      }
      if (!alignAlgo.matches("^bwasw|bwamem$")) {
        throw new ShearUsageException("Invalid alignment algorithm specified: " + alignAlgo);
      }

      // Validate origBwaIndexPrefix
      if (origBwaIndexPrefix == null) {
        origBwaIndexPrefix = origFaRefFile;
      }
      if (!new File(origBwaIndexPrefix + ".bwt").exists()) {
        throw new ShearUsageException("BWA index prefix appears to be invalid: " + origBwaIndexPrefix);
      } else {
        origBwaIndexPrefix = new File(origBwaIndexPrefix + ".bwt").getCanonicalPath();
        origBwaIndexPrefix = origBwaIndexPrefix.substring(0, origBwaIndexPrefix.length() - 4);
      }

      // Validate origTwoBitRefFile
      if (svAlgo.equals("crest")) {
        if (origTwoBitRefFile == null) {
          throw new ShearUsageException("When CREST is specified for the SV prediction algorithm, "
              + "you must provide a Two-Bit reference file with the '-t' argument");
        }
        if (!new File(origTwoBitRefFile).exists()) {
          throw new ShearUsageException("Two-Bit reference file does not exist: " + origTwoBitRefFile);
        } else {
          origTwoBitRefFile = new File(origTwoBitRefFile).getCanonicalPath();
        }
      } else {
        origTwoBitRefFile = null;
      }

      // Validate region
      if (region != null && !region.matches("^[^:]+(:[0-9,]+\\-[0-9,]+)?$")) {
        throw new ShearUsageException("Invalid region specified: " + region);
      }

      // Validate origPreds
      if (origPreds != null) {
        if (!new File(origPreds).exists()) {
          throw new ShearUsageException("SV prediction file does not exist: " + origPreds);
        } else {
          origPreds = new File(origPreds).getCanonicalPath();
        }
      }

      // Validate minHetThresh
      if (minHetThreshString != null) {
        try {
          minHetThresh = Double.parseDouble(minHetThreshString);
          if (minHetThresh < 0d || minHetThresh > 1d) {
            throw new ShearUsageException("Invalid minimum heterogeneity level specified: " + minHetThreshString);
          }
        } catch (NumberFormatException e) {
          throw new ShearUsageException("Invalid minimum heterogeneity level specified: " + minHetThreshString);
        }
      }

      // Validate crestDir
      if (svAlgo.equals("crest")) {
        if (crestDir == null) {
          throw new ShearUsageException("When CREST is specified for the SV prediction algorithm, "
              + "you must specify the CREST directory via the CREST_DIRECTORY environment variable");
        }
        if (!new File(crestDir).exists()) {
          throw new ShearUsageException("CREST directory does not exist: " + crestDir);
        } else {
          crestDir = new File(crestDir).getCanonicalPath();
        }
        if (!new File(crestDir + "/extractSClip.pl").exists()) {
          throw new ShearUsageException("CREST directory does not contain the necessary script extractSClip.pl: "
              + crestDir);
        }
        if (!new File(crestDir + "/CREST.pl").exists()) {
          throw new ShearUsageException("CREST directory does not contain the necessary script CREST.pl: " + crestDir);
        }
      } else {
        crestDir = null;
      }

      // Validate picardDir
      if (picardDir == null) {
        throw new ShearUsageException(
            "Must specify the Picard Tools directory via the PICARD_DIRECTORY environment variable");
      }
      if (!new File(picardDir).exists()) {
        throw new ShearUsageException("Picard Tools directory does not exist: " + picardDir);
      } else {
        picardDir = new File(picardDir).getCanonicalPath();
      }
      if (!(new File(picardDir + "/picard.jar").exists() || new File(picardDir + "/CleanSam.jar").exists()
          && new File(picardDir + "/FixMateInformation.jar").exists()
          && new File(picardDir + "/AddOrReplaceReadGroups.jar").exists())) {
        throw new ShearUsageException("Picard Tools directory (" + picardDir
            + ") does not contain the necessary JAR file(s), " + "must contain either picard.jar, or all of "
            + "CleanSam.jar, FixMateInformation.jar, and AddOrReplaceReadGroups.jar");
      }

      // Initialize all read group information to be the same as the prefix
      rgId = prefix;
      rgLb = prefix;
      rgPl = prefix;
      rgPu = prefix;
      rgSm = prefix;

      // Initialize read length to be 0.
      readLength = 0;

      // Initialize Genome object
      gen = new Genome(origFaRefFile, origFaRefFile + ".fai");

      // Handle errors
    } catch (ParseException e) {
      throw new ShearRuntimeException("Illegal format in specified *.fa or *.fa.fai file", e);
    } catch (IOException e) {
      throw new ShearRuntimeException("Problem occurred setting up SHEAR instance", e);
    }

    // Output settings information
    Shear.log("Starting SHEAR-SV (version " + Shear.VERSION + ")", false);
    Shear.log("Prefix:                  " + prefix, false);
    Shear.log("BAM alignment file:      " + origBamFile, false);
    Shear.log("FASTA reference file:    " + origFaRefFile, false);
    Shear.log("SV prediction algorithm: " + svAlgo, false);
    Shear.log("Alignment algorithm:     " + alignAlgo, false);
    if (origTwoBitRefFile != null) {
      Shear.log("Two-Bit reference file:  " + origTwoBitRefFile, false);
    }
    Shear.log("BWA index prefix:        " + origBwaIndexPrefix, false);
    if (region != null) {
      Shear.log("Region:                  " + region, false);
    }
    if (origPreds != null) {
      Shear.log("Orig predictions:        " + origPreds, false);
    }
    if (minHetThresh > 0d) {
      Shear.log(String.format("Minimum heterogeneity:   %.2f%%\n", 100d * minHetThresh), false);
    } else {
      Shear.log("Minimum heterogeneity:   N/A", false);
    }
    if (preprocess) {
      Shear.log("Preprocess input BAM:    YES", false);
    } else {
      Shear.log("Preprocess input BAM:    NO", false);
    }
    if (svOnly) {
      Shear.log("Prediction mode:         SV Only", false);
    } else {
      Shear.log("Prediction mode:         All", false);
    }
    if (realignUnmapped) {
      Shear.log("Realign unmapped reads:  ON", false);
    } else {
      Shear.log("Realign unmapped reads:  OFF", false);
    }
    if (debug) {
      Shear.log("Debugging mode:          ON", false);
    } else {
      Shear.log("Debugging mode:          OFF", false);
    }
    if (hetOnly) {
      Shear.log("Heterogeneity only:      ON", false);
    }
    Shear.log("Picard Tools directory:  " + picardDir, false);
    if (crestDir != null) {
      Shear.log("CREST directory:         " + crestDir, false);
    }

  }

  /**
   * Returns a string to use as a prefix for naming temporary files.
   *
   * @return A string to use as a prefix for naming temporary files
   */
  public String getTmpFilePrefix() {
    return prefix + ".tmp" + System.currentTimeMillis();
  }

  /**
   * Returns true if debug mode is set, false otherwise.
   *
   * @return True if debug mode is set, false otherwise
   */
  public boolean isDebugMode() {
    return debug;
  }

  /**
   * Scans the headers of a specified BAM file to determine read group
   * information and set it for this instance of ShearSV. If any read group
   * fields are not found, then they are set to the value of the read group ID.
   * If the read group ID is not found, then all are set to the value of the
   * prefix for this instance of ShearSV.
   *
   * @param bamFile
   *          The path to the BAM file to pull read group information from
   */
  private void determineReadGroupInfo(String bamFile) {

    // Set up scanner on string containing the BAM file headers
    Scanner sc = new Scanner(RunSystemCommand.run(new ProcessBuilder("samtools", "view", "-H", bamFile)));

    // Whether or not various read group information settings were found in headers
    boolean rgIdFound = false;
    boolean rgLbFound = false;
    boolean rgPlFound = false;
    boolean rgPuFound = false;
    boolean rgSmFound = false;

    // Loop through BAM file headers looking for the '@RG' line, then populate read group information instance variables
    while (sc.hasNextLine()) {
      String line = sc.nextLine();
      if (line.startsWith("@RG")) {
        String[] tokens = line.split("\t");
        for (int i = 1; i < tokens.length; i++) {
          if (tokens[i].startsWith("ID")) {
            rgId = tokens[i].substring(3);
            rgIdFound = true;
          } else if (tokens[i].startsWith("LB")) {
            rgLb = tokens[i].substring(3);
            rgLbFound = true;
          } else if (tokens[i].startsWith("PL")) {
            rgPl = tokens[i].substring(3);
            rgPlFound = true;
          } else if (tokens[i].startsWith("PU")) {
            rgPu = tokens[i].substring(3);
            rgPuFound = true;
          } else if (tokens[i].startsWith("SM")) {
            rgSm = tokens[i].substring(3);
            rgSmFound = true;
          }
        }
        if (!rgIdFound) {
          rgId = prefix;
        }
        if (!rgLbFound) {
          rgLb = rgId;
        }
        if (!rgPlFound) {
          rgPl = rgId;
        }
        if (!rgPuFound) {
          rgPu = rgId;
        }
        if (!rgSmFound) {
          rgSm = rgId;
        }
        break;
      }
    }

    // Close scanner
    sc.close();

  }

  /**
   * Scans the specified BAM file to determine the average read length and set
   * it for this instance of ShearSV.
   *
   * @param bamFile
   *          The path to the BAM file to pull read length information from
   */
  private void determineReadLength(String bamFile) {

    try {

      // GATK log mode
      String gatkLogMode;
      if (debug) {
        gatkLogMode = "INFO";
      } else {
        gatkLogMode = "ERROR";
      }

      // Maximum number of reads to scan
      int maxReadsToScan = 1000000;

      // File name for the temporary BAM/SAM files
      String tmpOutputPrefix = getTmpFilePrefix() + ".read-length-check";
      String tmpBamOutputFile = tmpOutputPrefix + ".bam";
      String tmpBamIndexOutputFile = tmpOutputPrefix + ".bai";
      String tmpSamOutputFile = tmpOutputPrefix + ".sam";

      // Write BAM file to temporary SAM file
      String gatkArgs =
          String.format(
              "-T PrintReads -I %s -R %s -n %d --read_filter NotPrimaryAlignment "
                  + "--maxRuntime 5 --maxRuntimeUnits MINUTES -o %s -l %s",
              bamFile,
              gen.getFaFilePath(),
              maxReadsToScan,
              tmpBamOutputFile,
              gatkLogMode);
      RunGatkCommand.run(gatkArgs.split(" "), debug);

      // Write temporary BAM file to temporary SAM file
      FileOutputStream output = new FileOutputStream(tmpSamOutputFile);
      RunSystemCommand.run(new ProcessBuilder("samtools", "view", tmpBamOutputFile), output, "samtools", debug);
      output.close();

      // HashMap to keep track of the counts of different read lengths
      HashMap<Integer, Long> counts = new HashMap<Integer, Long>();

      // Total number of reads in file
      long numReads = 0l;

      // Loop through temporary SAM file, storing counts of different read lengths
      Scanner sc = new Scanner(new File(tmpSamOutputFile));
      while (sc.hasNextLine()) {
        String[] samTokens = sc.nextLine().split("\t");
        int currentReadLength = samTokens[9].length();
        if (counts.containsKey(currentReadLength)) {
          counts.put(currentReadLength, counts.get(currentReadLength) + 1l);
        } else {
          counts.put(currentReadLength, 1l);
        }
        numReads++;
      }
      sc.close();

      // Delete temporary BAM/SAM files
      new File(tmpBamOutputFile).delete();
      new File(tmpBamIndexOutputFile).delete();
      new File(tmpSamOutputFile).delete();

      // Calculate overall average read length and set
      double doubleTypeReadLength = 0d;
      for (Map.Entry<Integer, Long> entry : counts.entrySet()) {
        doubleTypeReadLength += (double) entry.getValue() * (double) entry.getKey() / numReads;
      }
      readLength = (int) Math.round(doubleTypeReadLength);

      // Catch errors
    } catch (FileNotFoundException e) {
      throw new ShearRuntimeException("Unable to determine average read length in specified BAM file", e);
    } catch (IOException e) {
      throw new ShearRuntimeException("Unable to determine average read length in specified BAM file", e);
    }

  }

  /**
   * This prepares the original BAM file for SHEAR to use. An input BAM file is
   * specified, as well as an optional region to extract from the input BAM file
   * (can be null to use whole input BAM file). If preprocessing is required,
   * then the BAM file is cleaned, mate information is fixed, read group
   * information is added, and the BAM file is sorted. Returns the file path to
   * the processing BAM file (can be the same as in input BAM file path if no
   * region extraction nor preprocessing is done).
   *
   * @param inputBamFilePath
   *          Input BAM file to use
   * @param region
   *          Region to extract from the input BAM file and use as the output
   *          BAM file (can be null)
   * @returns The file path to the processed BAM file (can be the same as in
   *          input BAM file path if no region extraction nor preprocessing is
   *          done)
   */
  private String prepareOrigBamFile(String inputBamFilePath, String region) {

    // Name of current version of BAM file (updated as preprocessing is done)
    String currentBamFilePath = inputBamFilePath;

    // Files to use
    String regionOnlyBamFilePath = prefix + ".regionExtracted.bam";
    String preprocessedBamFilePath = prefix + ".preprocessed.bam";

    // Extract reads for specific region, if specified
    if (region != null) {
      Shear.log("Extracting reads from region " + region + "...", debug);
      try {
        FileOutputStream output = new FileOutputStream(regionOnlyBamFilePath);
        RunSystemCommand.run(
            new ProcessBuilder("samtools", "view", "-b", currentBamFilePath, region),
            output,
            "samtools",
            debug);
        output.close();
      } catch (FileNotFoundException e) {
        throw new ShearRuntimeException("Error creating BAM file to extract region '" + region + "' to");
      } catch (IOException e) {
        throw new ShearRuntimeException("Error creating BAM file to extract region '" + region + "' to");
      }
      currentBamFilePath = regionOnlyBamFilePath;
      RunSystemCommand.run(
          new ProcessBuilder("samtools", "index", regionOnlyBamFilePath),
          System.out,
          "samtools",
          debug);
    }

    // If preprocessing is necessary, then clean, fix mate information, add read groups, sort, and index
    if (preprocess) {
      Shear.log("Preprocessing original BAM alignment file...", debug);
      String cleanedFixedBamFilePath = getTmpFilePrefix() + ".cleaned.fixed.bam";
      cleanFixAndSort(currentBamFilePath, cleanedFixedBamFilePath);
      addReadGroupsAndSort(cleanedFixedBamFilePath, preprocessedBamFilePath);
      new File(cleanedFixedBamFilePath).delete();
      RunSystemCommand.run(
          new ProcessBuilder("samtools", "index", preprocessedBamFilePath),
          System.out,
          "samtools",
          debug);
      currentBamFilePath = preprocessedBamFilePath;
    }

    // Delete regionOnlyBamFilePath if uncessary
    if (!currentBamFilePath.equals(regionOnlyBamFilePath)) {
      new File(regionOnlyBamFilePath).delete();
      new File(regionOnlyBamFilePath + ".bai").delete();
    }

    // Return path to processed BAM file
    return currentBamFilePath;

  }

  /**
   * This takes a BAM/SAM file and creates a new BAM/SAM file that is cleaned,
   * has mate information fixed, and is sorted by coordinate.
   *
   * @param inputFilePath
   *          Input BAM/SAM file to use
   * @param outputFilePath
   *          Output BAM/SAM file to create that is cleaned, has fixed mate
   *          information, and is sorted
   */
  private void cleanFixAndSort(String inputFilePath, String outputFilePath) {

    // Temporary BAM file
    String tmpBamFile = getTmpFilePrefix() + ".cleaned.bam";

    // Clean BAM file
    String[] cleanSamArgs = { "I=" + inputFilePath, "O=" + tmpBamFile, "QUIET=TRUE", "VERBOSITY=ERROR" };
    if (debug) {
      cleanSamArgs[cleanSamArgs.length - 2] = "QUIET=FALSE";
      cleanSamArgs[cleanSamArgs.length - 1] = "VERBOSITY=INFO";
    }
    RunPicardCommand.run(picardDir, "CleanSam", cleanSamArgs, debug);

    // Fix mate information for BAM file
    String[] fixMateInformationArgs =
        { "I=" + tmpBamFile, "O=" + outputFilePath, "SO=coordinate", "QUIET=TRUE", "VERBOSITY=ERROR" };
    if (debug) {
      fixMateInformationArgs[fixMateInformationArgs.length - 2] = "QUIET=FALSE";
      fixMateInformationArgs[fixMateInformationArgs.length - 1] = "VERBOSITY=INFO";
    }
    RunPicardCommand.run(picardDir, "FixMateInformation", fixMateInformationArgs, debug);

    // Delete temporary file
    new File(tmpBamFile).delete();

  }

  /**
   * This takes a BAM/SAM file and creates a new BAM/SAM file that is sorted by
   * coordinate and has read group information added. The read group information
   * is taken from whatever the currently-set instance variables are for this
   * instance of ShearSV.
   *
   * @param inputFilePath
   *          Input BAM/SAM file to use
   * @param outputFilePath
   *          Output BAM/SAM file to create that is sorted and has read group
   *          information added
   */
  private void addReadGroupsAndSort(String inputFilePath, String outputFilePath) {
    String[] addOrReplaceReadGroupsArgs =
        { "I=" + inputFilePath, "O=" + outputFilePath, "SO=coordinate", "RGID=" + rgId, "RGLB=" + rgLb, "RGPL=" + rgPl,
            "RGPU=" + rgPu, "RGSM=" + rgSm, "QUIET=TRUE", "VERBOSITY=ERROR" };
    if (debug) {
      addOrReplaceReadGroupsArgs[addOrReplaceReadGroupsArgs.length - 2] = "QUIET=FALSE";
      addOrReplaceReadGroupsArgs[addOrReplaceReadGroupsArgs.length - 1] = "VERBOSITY=INFO";
    }
    RunPicardCommand.run(picardDir, "AddOrReplaceReadGroups", addOrReplaceReadGroupsArgs, debug);
  }

  /**
   * This takes a SAM file and creates a new BAM file that has MD tags fixed.
   * The input file must be SAM, and the output file will be BAM.
   *
   * @param inputSamFilePath
   *          Input SAM file to use
   * @param outputBamFilePath
   *          Output BAM file to create that has MD tags fixed
   */
  private void runCalmd(String inputSamFilePath, String outputBamFilePath) {
    try {
      FileOutputStream calmdOutput = new FileOutputStream(outputBamFilePath);
      RunSystemCommand.run(
          new ProcessBuilder("samtools", "calmd", "-S", "-b", inputSamFilePath, gen.getFaFilePath()),
          calmdOutput,
          "samtools",
          debug);
      calmdOutput.close();
    } catch (FileNotFoundException e) {
      throw new ShearRuntimeException("Error while running calmd on alignment file", e);
    } catch (IOException e) {
      throw new ShearRuntimeException("Error while running calmd on alignment file", e);
    }
  }

  /**
   * Takes a set of SV predictions and a set of SNP/INDEL predictions and
   * returns a combined set of filtered variants. The heterogeneity is assumed
   * to be set for both the SVs and the SNPs/INDELs.
   *
   * @param svs
   *          The set of SVs that have been predicted (must have heterogeneity
   *          estimates set already)
   * @param snpsIndels
   *          The setof SNPs and INDELs that have been predicted (must have
   *          heterogeneity estimates set already)
   * @returns The combined set of filtered variants
   */
  private TreeSet<Variant> filterVariants(TreeSet<Variant> svs, TreeSet<Variant> snpsIndels) {

    // Filter SV results
    Shear.log("Filtering SVs...", debug);
    svs = filterSvs(svs);

    // Make a new combined variant set, combinedVariants, that removes SNPs/INDELs adjacent to SV breakpoints
    Shear.log("Filtering SNPs/INDELs overlapping with SV breakpoints...", debug);
    TreeSet<Variant> combinedVariants = filterSnpsIndels(svs, snpsIndels);

    // Filter variants below minimum heterogeneity threshhold
    TreeSet<Variant> finalVariants = new TreeSet<Variant>();
    if (minHetThresh > 0d) {
      Shear.log(
          String.format("Filtering variants below %.2f%% estimated heterogeneity level...", 100d * minHetThresh),
          debug);
    }
    for (Variant v : combinedVariants) {
      if (v.getHeterogeneity() >= minHetThresh) {
        finalVariants.add(v);
      } else {
        Shear.log(
            String.format("  %s (%.2f%% is below minimum heterogeneity threshhold)", v, v.getHeterogeneity() * 100d),
            debug);
      }
    }

    // Return final filtered variant set
    return finalVariants;

  }

  /**
   * Filters a set of SVs to remove SVs that are near matches with another SV
   * but that have a lower heterogeneity level. A near match is defined as two
   * SVs of the same type (deletion, inversion, tandem duplication, or
   * translocation) that have one identical breakpoint with the other
   * corresponding breakpoints within 5bp of each other. This will remove false
   * positive SVs that have a slightly different breakpoint on one end with far
   * fewer soft-clipped reads (and thus a lower heterogeneity level).
   * Translocations that are equivalent to half of an inversion that is also
   * present are also removed. This also removes SVs that have a 0%
   * heterogeneity level.
   *
   * @param unfilteredSvs
   *          The set of SVs to filter
   * @return The filtered set of SVs
   */
  private TreeSet<Variant> filterSvs(TreeSet<Variant> unfilteredSvs) {

    // List of SVs sorted in descending order of heterogeneity (then sorted normally)
    ArrayList<Variant> hetSortedSvs = new ArrayList<Variant>();

    // Sort SVs in descending order of heterogeneity (then sort normally)
    tryNextSv: for (Variant sv : unfilteredSvs) {
      for (int i = 0; i < hetSortedSvs.size(); i++) {
        if (sv.getHeterogeneity() > hetSortedSvs.get(i).getHeterogeneity()
            || sv.getHeterogeneity() == hetSortedSvs.get(i).getHeterogeneity() && sv.compareTo(hetSortedSvs.get(i)) < 0) {
          hetSortedSvs.add(i, sv);
          continue tryNextSv;
        }
      }
      hetSortedSvs.add(sv);
    }

    // Arrays marking filter status and reason for all SVs
    ArrayList<Boolean> zeroHetFilter = new ArrayList<Boolean>();
    ArrayList<Integer> nearMatchFalsePositiveFilter = new ArrayList<Integer>();
    ArrayList<Integer> xlocInvFalsePositiveFilter = new ArrayList<Integer>();
    for (int i = 0; i < hetSortedSvs.size(); i++) {
      zeroHetFilter.add(false);
      nearMatchFalsePositiveFilter.add(-1);
      xlocInvFalsePositiveFilter.add(-1);
    }

    // Loop through SVs to filter for various reasons
    for (int i = 0; i < hetSortedSvs.size(); i++) {
      Variant outerLoopSv = hetSortedSvs.get(i);
      List<Breakpoint> outerLoopSvBps = outerLoopSv.getBreakpoints();

      // Check if SV has 0% heterogeneity
      if (outerLoopSv.getHeterogeneity() == 0d) {
        zeroHetFilter.set(i, true);
      }

      // Inner SV loop
      for (int j = 0; j < hetSortedSvs.size(); j++) {
        Variant innerLoopSv = hetSortedSvs.get(j);
        List<Breakpoint> innerLoopSvBps = innerLoopSv.getBreakpoints();
        if (!outerLoopSv.equals(innerLoopSv)) {

          // Check for near-matches where outerLoopSv has higher heterogeneity (or otherwise listed first)
          // Mark innerLoopSv as filtered due to outerLoopSv
          if (i < j && nearMatchFalsePositiveFilter.get(j) == -1) {

            // Check for deletion / tandem duplication / inversion near-match
            if (outerLoopSv instanceof Deletion && innerLoopSv instanceof Deletion
                || outerLoopSv instanceof TandemDuplication && innerLoopSv instanceof TandemDuplication
                || outerLoopSv instanceof Inversion && innerLoopSv instanceof Inversion) {

              // Calculate boolean values to use
              boolean bp1overlap = outerLoopSvBps.get(0).overlapsOrAdjacentTo(innerLoopSvBps.get(0), 0);
              boolean bp2overlap = outerLoopSvBps.get(1).overlapsOrAdjacentTo(innerLoopSvBps.get(1), 0);
              boolean bp1within5bp = outerLoopSvBps.get(0).overlapsOrAdjacentTo(innerLoopSvBps.get(0), 5);
              boolean bp2within5bp = outerLoopSvBps.get(1).overlapsOrAdjacentTo(innerLoopSvBps.get(1), 5);

              // Look for one identical breakpoint and second breakpoint ranges within 5bp of each other
              if (bp1overlap && bp2within5bp) {
                nearMatchFalsePositiveFilter.set(j, i);
              } else if (bp2overlap && bp1within5bp) {
                nearMatchFalsePositiveFilter.set(j, i);
              }

              // Check for translocation near-match
            } else if (outerLoopSv instanceof Translocation && innerLoopSv instanceof Translocation) {
              Translocation outerLoopXloc = (Translocation) outerLoopSv;
              Translocation innerLoopXloc = (Translocation) innerLoopSv;
              if (outerLoopXloc.getTransType() == innerLoopXloc.getTransType()) {

                // Calculate boolean values to use
                boolean bp1overlap = outerLoopSvBps.get(0).overlapsOrAdjacentTo(innerLoopSvBps.get(0), 0);
                boolean bp2overlap = outerLoopSvBps.get(1).overlapsOrAdjacentTo(innerLoopSvBps.get(1), 0);
                boolean bp1within5bp = outerLoopSvBps.get(0).overlapsOrAdjacentTo(innerLoopSvBps.get(0), 5);
                boolean bp2within5bp = outerLoopSvBps.get(1).overlapsOrAdjacentTo(innerLoopSvBps.get(1), 5);

                // Look for one identical breakpoint and second breakpoint ranges within 5bp of each other
                if (bp1overlap && bp2within5bp) {
                  nearMatchFalsePositiveFilter.set(j, i);
                } else if (bp2overlap && bp1within5bp) {
                  nearMatchFalsePositiveFilter.set(j, i);
                }

              }
            }

          }

          // Check for situations where innerLoopSv is a translocation that "belongs" to outLoopSv inversion
          // Mark innerLoopSv as filtered due to outerLoopSv
          if (outerLoopSv instanceof Inversion && innerLoopSv instanceof Translocation
              && xlocInvFalsePositiveFilter.get(j) == -1) {
            Inversion outerLoopInv = (Inversion) outerLoopSv;
            Translocation innerLoopXloc = (Translocation) innerLoopSv;
            if (innerLoopXloc.getChr().equals(outerLoopInv.getChr())
                && innerLoopXloc.getSecondChr().equals(outerLoopInv.getChr())) {
              if (innerLoopXloc.getTransType() == Translocation.Type.for3toRev5
                  && outerLoopInv.getBreakpoints().get(0)
                      .overlapsOrAdjacentTo(innerLoopXloc.getBreakpoints().get(0), 0)
                  && outerLoopInv.getBreakpoints().get(1)
                      .overlapsOrAdjacentTo(innerLoopXloc.getBreakpoints().get(1), 0)) {
                xlocInvFalsePositiveFilter.set(j, i);
              } else if (innerLoopXloc.getTransType() == Translocation.Type.for5toRev3
                  && outerLoopInv.getBreakpoints().get(0)
                      .overlapsOrAdjacentTo(innerLoopXloc.getBreakpoints().get(0), 0)
                  && outerLoopInv.getBreakpoints().get(1)
                      .overlapsOrAdjacentTo(innerLoopXloc.getBreakpoints().get(1), 0)) {
                xlocInvFalsePositiveFilter.set(j, i);
              }
            }
          }

        }
      }

    }

    // Set of SVs to return
    TreeSet<Variant> filteredSvs = new TreeSet<Variant>();

    // Loop through original SV set in order to display filter messages in correct order
    // Add unfiltered SVs to the set to return
    processNextSv: for (Variant sv : unfilteredSvs) {
      for (int i = 0; i < hetSortedSvs.size(); i++) {
        if (sv.equals(hetSortedSvs.get(i))) {
          if (zeroHetFilter.get(i)) {
            Shear.log("  " + sv + " (0% heterogeneity, considered false positive)", debug);
          } else if (nearMatchFalsePositiveFilter.get(i) != -1) {
            int idxOfFalsePositiveReason = nearMatchFalsePositiveFilter.get(i);
            while (nearMatchFalsePositiveFilter.get(idxOfFalsePositiveReason) != -1) {
              idxOfFalsePositiveReason = nearMatchFalsePositiveFilter.get(idxOfFalsePositiveReason);
            }
            Shear
                .log("  " + sv + " (false positive of SV '" + hetSortedSvs.get(idxOfFalsePositiveReason) + "')", debug);
          } else if (xlocInvFalsePositiveFilter.get(i) != -1) {
            Shear.log("  " + sv + " (false positive of SV '" + hetSortedSvs.get(xlocInvFalsePositiveFilter.get(i))
                + "')", debug);
          } else {
            filteredSvs.add(sv);
          }
          continue processNextSv;
        }
      }
    }

    // Return
    return filteredSvs;

  }

  /**
   * Filters a set of SNPs/INDELs to remove all that overlap or border a variant
   * in a set of SVs. This returns a combined set of the SVs and the filtered
   * SNPs/INDELs.
   *
   * @param svs
   *          The set of SVs to use
   * @param snpsIndels
   *          The set of SNPs and INDELs to filter
   * @return The combined set of SVs and the filtered SNPs/INDELs
   */
  private TreeSet<Variant> filterSnpsIndels(TreeSet<Variant> svs, TreeSet<Variant> snpsIndels) {

    // Set of combined variants to return
    TreeSet<Variant> combinedVariants = new TreeSet<Variant>();

    // Loop through SNPs/INDELs/
    for (Variant v : snpsIndels) {

      // Validate that set only contains SNPs/INDELs
      if (!(v instanceof SNP) && !v.isIndel()) {
        throw new ShearRuntimeException("Variants in SNPs/INDELs list must be SNPS or INDELs");
      }

      // Find the window of the SNP/INDEL (i.e. all affected bases)
      Chromosome snpIndelWindowChr = v.getChr();
      long snpIndelWindowStart = 0l;
      long snpIndelWindowEnd = 0l;
      if (v instanceof SNP) {
        snpIndelWindowStart = v.getCoord();
        snpIndelWindowEnd = v.getCoord();
      } else if (v instanceof Deletion) {
        snpIndelWindowStart = v.getBreakpoints().get(0).getFivePrimeCoord();
        snpIndelWindowEnd =
            Math.min(v.getBreakpoints().get(1).getChr().getLength(), v.getBreakpoints().get(1).getThreePrimeCoord() + 1);
      } else if (v instanceof Insertion) {
        snpIndelWindowStart = v.getBreakpoints().get(0).getFivePrimeCoord();
        snpIndelWindowEnd =
            Math.min(v.getBreakpoints().get(0).getChr().getLength(), v.getBreakpoints().get(0).getThreePrimeCoord() + 1);
      } else {
        throw new ShearRuntimeException("Variants in SNPs/INDELs list must be SNPS or INDELs");
      }

      // Keep track of SV that SNP/INDEL is adjacent to
      Variant adjacentSv = null;

      // Loop through SVs to check if SNP/INDEL to any SV breakpoints on same chromosome
      boolean overlaps = false;
      for (Variant sv : svs) {

        // Validate that set only contains SVs
        if (sv instanceof SNP || sv.isIndel()) {
          throw new ShearRuntimeException("Variants in SVs list must be SVs");
        }

        // Get breakpoints for SV
        List<Breakpoint> bps = sv.getBreakpoints();

        // Check if SNP/INDEL window is adjacent to any breakpoints
        for (Breakpoint bp : bps) {
          Chromosome svChr = sv.getChr();
          long bpWindowStart = bp.getFivePrimeCoord();
          long bpWindowEnd = Math.min(bp.getChr().getLength(), bp.getThreePrimeCoord() + 1);
          if (svChr.equals(snpIndelWindowChr)
              && Math.max(bpWindowStart, snpIndelWindowStart) <= Math.min(bpWindowEnd, snpIndelWindowEnd)) {
            overlaps = true;
            adjacentSv = sv;
            break;
          }
        }

      }

      // If SNP/INDEL doesn't overlap with an SV breakpoint, then add it to combined list
      if (!overlaps) {
        combinedVariants.add(v);
      } else {
        Shear.log("  " + v + " (SNP/INDEL that bordered breakpoint of SV '" + adjacentSv + "')", debug);
      }

    }

    // Add all SVs to combined variant list
    combinedVariants.addAll(svs);

    // Return combined variant set
    return combinedVariants;

  }

  /**
   * Outputs the set of variants to SDI and report files. If SV-Only mode is on,
   * then only *.sv.sdi and *.sv.report are created. Otherwise, all of *.sv.sdi,
   * *.sv.report, *.snpindel.sdi, *.snpindel.report, *.all.sdi, and
   * *.all.snpindel are created.
   *
   * @param variants
   *          The set of variants to output
   */
  private void outputResults(TreeSet<Variant> variants) {

    // Keep track of if we found SNPs/INDELs or SVs
    boolean foundSvs = false;
    boolean foundSnpIndels = false;

    // Array to store lines to output to report file
    ArrayList<ReportLine> reportLines = new ArrayList<ReportLine>();

    // Array to keep track of SNPs/INDELs vs. SVs
    ArrayList<Boolean> lineIsSv = new ArrayList<Boolean>();

    // Initialize output streams to null
    PrintWriter allSdiOut = null;
    PrintWriter allReportOut = null;
    PrintWriter snpIndelSdiOut = null;
    PrintWriter snpIndelReportOut = null;
    PrintWriter svSdiOut = null;
    PrintWriter svReportOut = null;

    try {

      // Open output streams
      if (!svOnly) {
        allSdiOut = new PrintWriter(prefix + ".all.sdi");
        allReportOut = new PrintWriter(prefix + ".all.report");
        snpIndelSdiOut = new PrintWriter(prefix + ".snpindel.sdi");
        snpIndelReportOut = new PrintWriter(prefix + ".snpindel.report");
      }
      svSdiOut = new PrintWriter(prefix + ".sv.sdi");
      svReportOut = new PrintWriter(prefix + ".sv.report");

      // Loop through all variants
      for (Variant v : variants) {

        // Update foundSvs and foundSnpIndels booleans and add to lineIsSv array
        boolean currentLineIsSv;
        if (v instanceof SNP || v.isIndel()) {
          currentLineIsSv = false;
          foundSnpIndels = true;
        } else {
          currentLineIsSv = true;
          foundSvs = true;
        }
        lineIsSv.add(currentLineIsSv);

        // Print SDI line
        String sdiLine = v.getSdiString();
        if (sdiLine != null) {
          if (currentLineIsSv) {
            svSdiOut.println(sdiLine);
          }
          if (!svOnly && !currentLineIsSv) {
            snpIndelSdiOut.println(sdiLine);
          }
          if (!svOnly) {
            allSdiOut.println(sdiLine);
          }
        } else {
          Shear.log("  Translocation will be skipped for SDI output '" + v + "'", debug);
        }

        // Create report line for SNP
        if (v instanceof SNP) {
          SNP snp = (SNP) v;
          reportLines.add(new ReportLine(
              v.getVariantType(),
              v.getVariantDetails(),
              v.getBreakpointMicrohomology(),
              String.valueOf(snp.getRefBase()),
              String.valueOf(snp.getVarBase()),
              v.getHeterogeneity()));

          //  Create report line for deletion
        } else if (v instanceof Deletion) {
          Deletion del = (Deletion) v;
          reportLines.add(new ReportLine(v.getVariantType(), v.getVariantDetails(), v.getBreakpointMicrohomology(), del
              .getDeletedBases(), "-", v.getHeterogeneity()));

          // Create report line for insertion
        } else if (v instanceof Insertion) {
          Insertion ins = (Insertion) v;
          reportLines.add(new ReportLine(
              v.getVariantType(),
              v.getVariantDetails(),
              v.getBreakpointMicrohomology(),
              "-",
              ins.getInsertedBases(),
              v.getHeterogeneity()));

          //  Create report line for tandem duplication
        } else if (v instanceof TandemDuplication) {
          TandemDuplication tdup = (TandemDuplication) v;
          reportLines.add(new ReportLine(
              v.getVariantType(),
              v.getVariantDetails(),
              v.getBreakpointMicrohomology(),
              tdup.getDuplicatedBases(),
              tdup.getDuplicatedBases() + tdup.getDuplicatedBases(),
              v.getHeterogeneity()));

          // Create report line for inversion
        } else if (v instanceof Inversion) {
          Inversion inv = (Inversion) v;
          reportLines.add(new ReportLine(v.getVariantType(), v.getVariantDetails(), v.getBreakpointMicrohomology(), inv
              .getAffectedBases(), inv.getInvertedBases(), v.getHeterogeneity()));

          // Create report line for translocation
        } else if (v instanceof Translocation) {
          reportLines.add(new ReportLine(
              v.getVariantType(),
              v.getVariantDetails(),
              v.getBreakpointMicrohomology(),
              "-",
              "-",
              v.getHeterogeneity()));

          // Unknown variant type
        } else {
          throw new IllegalArgumentException("Encountered an invalid variant type while outputting results");
        }

      }

      // Determine longest string for loc/size in report output (minimum size of 21)
      int longestDetailsString = 21;
      for (ReportLine line : reportLines) {
        if (line.details.length() > longestDetailsString) {
          longestDetailsString = line.details.length();
        }
      }

      // Report header line format string
      String reportHeaderFormat = "%-6s%-" + longestDetailsString + "s  %13s  %-21s%-21s%13s";

      // Report line format string
      String reportFormat = "%-6s%-" + longestDetailsString + "s  %13s  %-21s%-21s%12.2f%%";

      // Header line to use for reports
      String reportHeaderLine =
          String.format(
              reportHeaderFormat,
              "Type",
              "Details",
              "Microhomology",
              "Reference Bases",
              "Variant Bases",
              "Heterogeneity");

      // Output report file headers (and to standard output)
      if (foundSvs) {
        svReportOut.println(reportHeaderLine);
      }
      if (!svOnly && foundSnpIndels) {
        snpIndelReportOut.println(reportHeaderLine);
      }
      if (!svOnly && (foundSvs || foundSnpIndels)) {
        allReportOut.println(reportHeaderLine);
      }
      if (foundSvs || foundSnpIndels) {
        Shear.log("  " + reportHeaderLine, debug);
      } else {
        Shear.log("  No results!", debug);
      }

      // Output report file lines (and to standard output)
      for (int i = 0; i < reportLines.size(); i++) {
        ReportLine rl = reportLines.get(i);
        String microStr = rl.microhomology > 0l ? String.format("%d bp", rl.microhomology) : "-";
        if (lineIsSv.get(i)) {
          svReportOut.printf(reportFormat, rl.type, rl.details, microStr, rl.refBases, rl.varBases, 100d * rl.het);
          svReportOut.println();
        }
        if (!svOnly && !lineIsSv.get(i)) {
          snpIndelReportOut
              .printf(reportFormat, rl.type, rl.details, microStr, rl.refBases, rl.varBases, 100d * rl.het);
          snpIndelReportOut.println();
        }
        if (!svOnly) {
          allReportOut.printf(reportFormat, rl.type, rl.details, microStr, rl.refBases, rl.varBases, 100d * rl.het);
          allReportOut.println();
        }
        Shear.log(
            "  " + String.format(reportFormat, rl.type, rl.details, microStr, rl.refBases, rl.varBases, 100d * rl.het),
            debug);
      }

      // Catch errors
    } catch (FileNotFoundException e) {
      throw new ShearRuntimeException("Unable to write to output files", e);

      // Close output streams
    } finally {
      if (!svOnly) {
        allSdiOut.close();
        allReportOut.close();
        snpIndelSdiOut.close();
        snpIndelReportOut.close();
      }
      svSdiOut.close();
      svReportOut.close();
    }

  }

  /**
   * ReportLine is a helper class to be used by the outputResults method and
   * store the information that appears on one line of the report file output
   * (i.e. one variant).
   */
  private class ReportLine {

    // The type of variant
    private String type;

    // The details of the variant
    private String details;

    // The length of the microhomology of the variant
    private long   microhomology;

    // The reference bases string (can be truncated)
    private String refBases;

    // The variant bases string (can be truncated)
    private String varBases;

    // The heterogeneity of the variant
    private double het;

    /**
     * Creates a new ReportLine object. If refBases or varBases are longer than
     * 19 characters, they will be truncated in the middle.
     *
     * @param type
     *          The type of variant
     * @param details
     *          The details of the variant
     * @param microhomology
     *          The length of the microhomology of the variant
     * @param refBases
     *          The reference bases string (can be truncated)
     * @param varBases
     *          The variant bases string (can be truncated)
     * @param het
     *          The heterogeneity of the variant
     */
    private ReportLine(String type, String details, long microhomology, String refBases, String varBases, double het) {

      // Truncate reference bases string
      if (refBases.length() > 19) {
        refBases = refBases.substring(0, 8) + "..." + refBases.substring(refBases.length() - 8);
      }

      // Truncate reference bases string
      if (varBases.length() > 19) {
        varBases = varBases.substring(0, 8) + "..." + varBases.substring(varBases.length() - 8);
      }

      // Set instance variables
      this.type = type;
      this.details = details;
      this.microhomology = microhomology;
      this.refBases = refBases;
      this.varBases = varBases;
      this.het = het;

    }

  }

  /**
   * Main method to run the ShearSV module.
   */
  public void execute() {

    // Determine the read group information
    Shear.log("Extracting read group information...", debug);
    determineReadGroupInfo(origBamFile);

    // Prepare BAM file for use
    String processedOrigBamFile = prepareOrigBamFile(origBamFile, region);

    // Determine the average read length
    Shear.log("Determining read lengths for BAM file...", debug);
    determineReadLength(processedOrigBamFile);

    // Initialize variant caller
    VariantCaller svCaller = null;
    if (svAlgo.equals("delly")) {
      svCaller = new DellyWrapper(gen, debug);
    } else if (svAlgo.equals("crest")) {
      svCaller = new CrestWrapper(gen, debug, readLength, crestDir, prefix, origTwoBitRefFile);
    }

    // If Heterogeneity Only mode is set, just get SVs, estimate heterogeneity, filter, and output
    if (hetOnly) {

      TreeSet<Variant> svs = svCaller.getVariants(origPreds);
      Shear.log("Estimating Heterogeneity for SVs...", debug);
      new SvHeterogeneityEstimator(processedOrigBamFile, origFaRefFile, this).estimateHeterogeneity(svs);
      TreeSet<Variant> finalVariants = filterVariants(svs, new TreeSet<Variant>());
      Shear.log("Outputting Results...", debug);
      outputResults(finalVariants);
      Shear.log("All Finished!", false);

    } else {

      // Predict SNPs/INDELs
      TreeSet<Variant> snpsIndels = new TreeSet<Variant>();
      if (!svOnly) {
        Shear.log("Predicting SNPs and INDELs...", debug);
        HaplotypeCallerWrapper haplotypeCaller = new HaplotypeCallerWrapper(gen, debug);
        String haplotypeCallerOutputFilePath =
            haplotypeCaller.runVariantCaller(processedOrigBamFile, prefix + ".snpindel");
        snpsIndels = haplotypeCaller.getVariants(haplotypeCallerOutputFilePath);
        if (!debug) {
          new File(haplotypeCallerOutputFilePath).delete();
          new File(haplotypeCallerOutputFilePath + ".idx").delete();
        }
        haplotypeCaller.close();
      }

      // Initial SV predictions
      String origSvPredsFile;
      if (origPreds == null) {
        Shear.log("Making SV Predictions with " + svAlgo.toUpperCase() + "...", debug);
        origSvPredsFile = svCaller.runVariantCaller(processedOrigBamFile, prefix + ".0");
      } else {
        origSvPredsFile = origPreds;
      }
      TreeSet<Variant> svs = svCaller.getVariants(origSvPredsFile);
      if (!debug && origPreds == null) {
        new File(origSvPredsFile).delete();
      }

      // Report variants from initial SV prediction and set hasNewVariants boolean
      Shear.log("Predicted SVs:", debug);
      boolean hasNewVariants = true;
      if (svs.size() > 0) {
        for (Variant sv : svs) {
          Shear.log("  " + sv, debug);
        }
      } else {
        Shear.log("  None", debug);
        hasNewVariants = false;
      }

      // Breakpoints realigner instance
      BreakpointsRealigner realigner =
          new BreakpointsRealigner(processedOrigBamFile, gen, origBwaIndexPrefix, alignAlgo, realignUnmapped, this);

      // Delete files that may be present from a previous run
      for (int i = 0; i <= 5; i++) {
        new File(prefix + "." + i + ".extracted.aligned.bam").delete();
        new File(prefix + "." + i + ".extracted.aligned.bam.bai").delete();
      }

      // Continue iterations until no new variants from SV predictions or we've gone 5 times
      // This needs to go at least one iteration in order to generate *.1.predSV.txt, which later code assumes
      int ctr = 0;
      String extractedBamFileName = null;
      while (ctr < 5 && hasNewVariants) {
        ctr++;
        String currentPrefix = prefix + "." + ctr;

        // Temporary files
        String tmpExtractedSamFileName = getTmpFilePrefix() + ".extracted.aligned.sam";
        String tmpExtractedCleanedFixedSamFileName = getTmpFilePrefix() + ".extracted.aligned.cleaned.fixed.sam";
        String tmpExtractedCleanedFixedRgAddedSamFileName =
            getTmpFilePrefix() + ".extracted.aligned.cleaned.fixed.rgadded.sam";

        // Extract relevant reads and realign
        Shear.log("Extracting and Realigning Breakpoint and Unmapped Reads...", debug);
        extractedBamFileName = currentPrefix + ".extracted.aligned.bam";
        realigner.extractAndRealign(svs, tmpExtractedSamFileName);

        // Clean, fix mate information, add read groups, and sort for new realignment
        // If alignment algorithm is BWA-SW, then also run calmd on final SAM file
        if (alignAlgo.equals("bwasw")) {
          cleanFixAndSort(tmpExtractedSamFileName, tmpExtractedCleanedFixedSamFileName);
          addReadGroupsAndSort(tmpExtractedCleanedFixedSamFileName, tmpExtractedCleanedFixedRgAddedSamFileName);
          runCalmd(tmpExtractedCleanedFixedRgAddedSamFileName, extractedBamFileName);
        } else {
          cleanFixAndSort(tmpExtractedSamFileName, tmpExtractedCleanedFixedSamFileName);
          addReadGroupsAndSort(tmpExtractedCleanedFixedSamFileName, extractedBamFileName);
        }

        // Index new realignment
        RunSystemCommand.run(
            new ProcessBuilder("samtools", "index", extractedBamFileName),
            System.out,
            "samtools",
            debug);

        // Clear temporary files
        new File(tmpExtractedSamFileName).delete();
        new File(tmpExtractedCleanedFixedSamFileName).delete();
        new File(tmpExtractedCleanedFixedRgAddedSamFileName).delete();

        // Make new SV predictions
        Shear.log("Making SV Predictions with " + svAlgo.toUpperCase() + "...", debug);
        ArrayList<String> svCAllerArgs = new ArrayList<String>();
        if (svAlgo.equals("crest")) {
          svCAllerArgs.add("--min_sclip_reads");
          svCAllerArgs.add("2");
        }
        String svPredsFile = svCaller.runVariantCaller(extractedBamFileName, currentPrefix, svCAllerArgs);
        TreeSet<Variant> newSvs = svCaller.getVariants(svPredsFile);
        if (!debug) {
          new File(svPredsFile).delete();
        }

        // Display new variants from new SV prediction, add them to set, and update hasNewVariants boolean
        hasNewVariants = false;
        for (Variant sv : newSvs) {
          if (!svs.contains(sv)) {
            Shear.log("  New variant: " + sv, debug);
            svs.add(sv);
            hasNewVariants = true;
          }
        }

      }

      // Estimate heterogeneity for all SVs
      if (ctr > 0) {
        Shear.log("  No new variants detected", debug);
        Shear.log("Estimating Heterogeneity for SVs...", debug);
        new SvHeterogeneityEstimator(extractedBamFileName, origFaRefFile, this).estimateHeterogeneity(svs);
      }

      // Filter variants
      TreeSet<Variant> finalVariants = filterVariants(svs, snpsIndels);

      // Output results
      Shear.log("Outputting Results...", debug);
      outputResults(finalVariants);

      // Display final log message
      Shear.log("All Finished!", false);

      // Cleanup files
      if (!debug) {
        for (int i = 0; i <= ctr; i++) {
          new File(prefix + "." + i + ".extracted.aligned.bam").delete();
          new File(prefix + "." + i + ".extracted.aligned.bam.bai").delete();
        }
      }

    }

    // Close Genome object
    try {
      gen.close();
    } catch (IOException e) {
      throw new ShearRuntimeException("Unexpected error while trying to close Genome object", e);
    }

    // Close SV Caller
    svCaller.close();

  }

}
