/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.aligners;

import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import edu.umn.cs.kumarbio.shear.ShearSV;
import edu.umn.cs.kumarbio.shear.system.RunSystemCommand;

/**
 * Represents an instance of a BWA-SW aligner. This class extends the abstract
 * class BwaAligner.
 */
public class BwaSwAligner extends BwaAligner {

  /**
   * Creates a new BwaSwAligner object for the specified reference file. The
   * generateBwaIndex parameter can be set to true to generate BWA index files
   * for this aligner instance to use. If generateBwaIndex is false, then the
   * proper BWA index files are expected to be present already with the
   * reference file path being the prefix for the BWA index files.
   *
   * @param faFile
   *          The path to the reference file to align to
   * @param generateBwaIndex
   *          Whether or not to generate BWA index files for this aligner
   *          instance
   * @param shearSvInstance
   *          The instance of ShearSV that this is called from
   */
  public BwaSwAligner(String faFilePath, boolean generateBwaIndex, ShearSV shearSvInstance) {
    super(faFilePath, generateBwaIndex, shearSvInstance);
  }

  @Override
  public void align(String fastqInputFilePath, String samOutputFilePath, List<String> options)
      throws FileNotFoundException, IOException {

    // Open SAM output
    FileOutputStream out = new FileOutputStream(samOutputFilePath);

    // Prepare list of command and arguments for BWA
    ArrayList<String> bwaCommandAndArgs = new ArrayList<String>();
    bwaCommandAndArgs.add("bwa");
    bwaCommandAndArgs.add("bwasw");
    bwaCommandAndArgs.add("-M");
    for (int i = 0; i < options.size(); i++) {
      bwaCommandAndArgs.add(options.get(i));
    }
    bwaCommandAndArgs.add(getFaFilePath());
    bwaCommandAndArgs.add(fastqInputFilePath);

    // Run BWA command
    RunSystemCommand.run(new ProcessBuilder(bwaCommandAndArgs), out, "BWA-SW", isDebugMode());

    // Close SAM output
    out.close();

  }

}
