/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.heterogeneity;

import edu.umn.cs.kumarbio.Chromosome;
import edu.umn.cs.kumarbio.shear.ShearRuntimeException;
import edu.umn.cs.kumarbio.shear.variants.Location;

/**
 * This represents a breakpoint for a variant.
 */
public class Breakpoint implements Comparable<Breakpoint> {

  // The 5'-most possible location for this breakpoint (i.e. coordinate just before the break)
  private Location        fivePrimeLoc;

  // The 3'-most possible location for this breakpoint (i.e. coordinate just before the break)
  private Location        threePrimeLoc;

  // Whether or not the 5'-most possible location for this breakpoint occurs before the start of the chromosome
  // (i.e. between coordinates 0 and 1)
  private boolean         fivePrimeLocIsBeforeStart;

  // Whether or not the 3'-most possible location for this breakpoint occurs before the start of the chromosome
  // (i.e. between coordinates 0 and 1)
  private boolean         threePrimeLocIsBeforeStart;

  // The information about counts of reads that are clipped at or span the breakpoint
  private BreakpointReads bpReads;

  /**
   * Creates a new Breakpoint object using the specified location range. The
   * actual breakpoint could occur anywhere between the two specified locations
   * due to microhomology. The location represents the coordinate just before
   * the break. If the break occurs before the start of the chromosome (i.e.
   * between coordinates 0 and 1), then the location coordinate must be 1 and
   * the relevant *IsBeforeStart argument must be true. Initializes the
   * breakpoint reads to zeros.
   *
   * @param fivePrimeLoc
   *          The 5'-most possible location for this breakpoint (i.e. coordinate
   *          just before the break), or the 1 coordinate in conjunction with
   *          fivePrimeLocIsBeforeStart=true to represent a break that occurs
   *          before the start of the chromosome (i.e. between coordinates 0 and
   *          1)
   * @param threePrimeLoc
   *          The 3'-most possible location for this breakpoint (i.e. coordinate
   *          just before the break), or the 1 coordinate in conjunction with
   *          threePrimeLocIsBeforeStart=true to represent a break that occurs
   *          before the start of the chromosome (i.e. between coordinates 0 and
   *          1)
   * @param fivePrimeLocIsBeforeStart
   *          Whether or not the 5'-most possible location for this breakpoint
   *          occurs before the start of the chromosome (i.e. between
   *          coordinates 0 and 1)
   * @param threePrimeLocIsBeforeStart
   *          Whether or not the 3'-most possible location for this breakpoint
   *          occurs before the start of the chromosome (i.e. between
   *          coordinates 0 and 1)
   */
  public Breakpoint(
      Location fivePrimeLoc, Location threePrimeLoc, boolean fivePrimeLocIsBeforeStart,
      boolean threePrimeLocIsBeforeStart) {

    // Validate arguments
    if (fivePrimeLoc == null || threePrimeLoc == null) {
      throw new ShearRuntimeException("Null location(s) specified for breakpoint");
    } else if (!fivePrimeLoc.getChr().equals(threePrimeLoc.getChr())) {
      throw new ShearRuntimeException("Possible locations for breakpoint must be on same chromosome");
    } else if (fivePrimeLoc.getCoord() > threePrimeLoc.getCoord()) {
      throw new ShearRuntimeException("3'-most possible location must be at least as high as 5'-most possible location");
    } else if (fivePrimeLoc.getCoord() == 1l && threePrimeLoc.getCoord() == 1l && !fivePrimeLocIsBeforeStart
        && threePrimeLocIsBeforeStart) {
      throw new ShearRuntimeException("3'-most possible location must be at least as high as 5'-most possible location");
    } else if (fivePrimeLocIsBeforeStart && fivePrimeLoc.getCoord() != 1l) {
      throw new ShearRuntimeException(
          "The fivePrimeLocIsBeforeStart flag can only be set if the 5'-most possible location coordinate is set to 1");
    } else if (threePrimeLocIsBeforeStart && threePrimeLoc.getCoord() != 1l) {
      throw new ShearRuntimeException(
          "The threePrimeLocIsBeforeStart flag can only be set if the 3'-most possible location coordinate is set to 1");
    }

    // Set instance variables
    this.fivePrimeLoc = fivePrimeLoc;
    this.threePrimeLoc = threePrimeLoc;
    this.fivePrimeLocIsBeforeStart = fivePrimeLocIsBeforeStart;
    this.threePrimeLocIsBeforeStart = threePrimeLocIsBeforeStart;
    bpReads = new BreakpointReads(0, 0, 0, 0);

  }

  /**
   * Returns the chromosome that this breakpoint occurs on.
   *
   * @return The chromosome that this breakpoint occurs on
   */
  public Chromosome getChr() {
    return fivePrimeLoc.getChr();
  }

  /**
   * Returns the coordinate of the 5'-most possible location for this breakpoint
   * (i.e. coordinate just before the break). This can be 0 to represent a break
   * that occurs before the start of the chromosome.
   *
   * @return The coordinate of the 5'-most possible location for this breakpoint
   *         (i.e. coordinate just before the break)
   */
  public long getFivePrimeCoord() {
    if (fivePrimeLocIsBeforeStart) {
      return 0;
    } else {
      return fivePrimeLoc.getCoord();
    }
  }

  /**
   * Returns the coordinate of the 3'-most possible location for this breakpoint
   * (i.e. coordinate just before the break). This can be 0 to represent a break
   * that occurs before the start of the chromosome.
   *
   * @return The coordinate of the 3'-most possible location for this breakpoint
   *         (i.e. coordinate just before the break)
   */
  public long getThreePrimeCoord() {
    if (threePrimeLocIsBeforeStart) {
      return 0;
    } else {
      return threePrimeLoc.getCoord();
    }
  }

  /**
   * Returns the information about counts of reads that are clipped at or span
   * the breakpoint.
   *
   * @return The information about counts of reads that are clipped at or span
   *         the breakpoint
   */
  public BreakpointReads getBpReads() {
    return bpReads;
  }

  /**
   * Returns true if the two breakpoints have possible locations that overlap
   * one another, or that are within the allowable separation to one another.
   * Use '0' for allowableSeparation to check if the two breakpoints overlap.
   *
   * @param other
   *          The breakpoint to compare to
   * @param allowableSeparation
   *          The maximum number of bases that the two breakpoint ranges can be
   *          separated by in order to return true.
   * @return True if the two breakpoints have possible locations that overlap
   *         one another, or that are within the allowable separation to one
   *         another
   */
  public boolean overlapsOrAdjacentTo(Breakpoint other, int allowableSeparation) {
    if (!getChr().equals(other.getChr())) {
      return false;
    } else {
      return Math.max(getFivePrimeCoord(), other.getFivePrimeCoord()) <= Math.min(
          getThreePrimeCoord(),
          other.getThreePrimeCoord())
          + allowableSeparation;
    }
  }

  @Override
  public int compareTo(Breakpoint other) {

    // Return 0 if breakpoints are equal
    if (equals(other)) {
      return 0;
    }

    // Return whichever has first chromosome
    if (getChr().compareTo(other.getChr()) < 0) {
      return -1;
    } else if (getChr().compareTo(other.getChr()) > 0) {
      return +1;
    }

    // Return whichever has first 5' coordinate
    if (getFivePrimeCoord() < other.getFivePrimeCoord()) {
      return -1;
    } else if (getFivePrimeCoord() > other.getFivePrimeCoord()) {
      return +1;
    }

    // Return whichever has first 3' coordinate
    if (getThreePrimeCoord() < other.getThreePrimeCoord()) {
      return -1;
    } else if (getThreePrimeCoord() > other.getThreePrimeCoord()) {
      return +1;
    }

    // If locations are the same, compare the strings of the BreakpointReads
    if (getBpReads().toString().compareTo(other.getBpReads().toString()) < 0) {
      return -1;
    } else if (getBpReads().toString().compareTo(other.getBpReads().toString()) > 0) {
      return +1;

      // Throw error for exact same BreakpointReads string (should be unreachable code)
    } else {
      throw new ShearRuntimeException("Unexpected behavior while comparing order of Breakpoint objects"
          + " (Same locations and breakpoint reads, but objects not equal)");
    }

  }

  @Override
  public boolean equals(Object obj) {
    if (obj == null || !(obj instanceof Breakpoint)) {
      return false;
    } else {
      Breakpoint other = (Breakpoint) obj;
      return getChr().equals(other.getChr()) && getFivePrimeCoord() == other.getFivePrimeCoord()
          && getThreePrimeCoord() == other.getThreePrimeCoord() & getBpReads().equals(other.getBpReads());
    }
  }

  @Override
  public String toString() {
    return String.format(
        "%s:%d (%dbp microhomology) (reads: %s)",
        getChr().getName(),
        getFivePrimeCoord(),
        getThreePrimeCoord() - getFivePrimeCoord(),
        getBpReads());
  }

}
