/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.heterogeneity;

import htsjdk.samtools.CigarElement;
import htsjdk.samtools.CigarOperator;

import java.io.PrintStream;

import org.broadinstitute.gatk.engine.CommandLineGATK;
import org.broadinstitute.gatk.engine.walkers.DataSource;
import org.broadinstitute.gatk.engine.walkers.ReadWalker;
import org.broadinstitute.gatk.engine.walkers.Requires;
import org.broadinstitute.gatk.utils.commandline.Argument;
import org.broadinstitute.gatk.utils.commandline.Output;
import org.broadinstitute.gatk.utils.contexts.ReferenceContext;
import org.broadinstitute.gatk.utils.help.DocumentedGATKFeature;
import org.broadinstitute.gatk.utils.refdata.RefMetaDataTracker;
import org.broadinstitute.gatk.utils.sam.GATKSAMRecord;

import edu.umn.cs.kumarbio.shear.ShearRuntimeException;

/**
 * BreakpointReadsWalker is an extension of GATK's ReadWalker that will count
 * the number of reads that span a given given versus those that are clipped in
 * a either direction at the given breakpoint. This handles breakpoint ambiguity
 * via microhomology by taking a range of possible breakpoint locations as input
 * and returning the number of clipped and spanning reads at the 5'-most and
 * 3'-most edges of possible breakpoint locations. Breakpoint locations are
 * specified as the coordinate just before the break (note that this can be 0 if
 * a break occurs before the beginning of a chromosome). This is used by the
 * setBreakpointReads method in the SvHeterogeneityEstimator class.
 */
@DocumentedGATKFeature(groupName = "Quality Control and Simple Analysis Tools", extraDocs = { CommandLineGATK.class })
@Requires({ DataSource.READS, DataSource.REFERENCE })
public class BreakpointReadsWalker extends ReadWalker<BreakpointReads, BreakpointReads> {

  @Output
  protected PrintStream out;

  @Argument(
      fullName = "fivePrimeEdgeBpLoc",
      shortName = "fivePrimeEdgeBpLoc",
      doc = "5'-most edge of possible breakpoint locations.",
      required = true)
  protected int         FIVE_PRIME_EDGE_BP_LOC  = -1;

  @Argument(
      fullName = "threePrimeEdgeBpLoc",
      shortName = "threePrimeEdgeBpLoc",
      doc = "3'-most edge of possible breakpoint locations.",
      required = true)
  protected int         THREE_PRIME_EDGE_BP_LOC = -1;

  @Argument(
      fullName = "unclippedSpanningOnly",
      shortName = "unclippedSpanningOnly",
      doc = "Only count spanning reads that are completely unclipped (i.e. not just at the breakpoint).",
      required = false)
  protected boolean     UNCLIPPED_SPANNING_ONLY = false;

  @Override
  public BreakpointReads reduceInit() {
    return new BreakpointReads(0, 0, 0, 0);
  }

  @Override
  public BreakpointReads reduce(BreakpointReads value, BreakpointReads sum) {
    sum.addToFivePrimeClippedReads(value.getFivePrimeClippedReads());
    sum.addToThreePrimeClippedReads(value.getThreePrimeClippedReads());
    sum.addToFivePrimeSpanningReads(value.getFivePrimeSpanningReads());
    sum.addToThreePrimeSpanningReads(value.getThreePrimeSpanningReads());
    return sum;
  }

  @Override
  public BreakpointReads map(ReferenceContext ref, GATKSAMRecord read, RefMetaDataTracker tracker) {

    // If the read is unmapped, then return 0/0/0/0
    if (read.getReadUnmappedFlag()) {
      return new BreakpointReads(0, 0, 0, 0);
    }

    // If the first CIGAR element is soft-clip, and alignment starts right after 5'-most breakpoint, then return 1/0/0/0
    if (read.getAlignmentStart() == FIVE_PRIME_EDGE_BP_LOC + 1
        && read.getCigar().getCigarElement(0).getOperator() == CigarOperator.S) {
      return new BreakpointReads(1, 0, 0, 0);
    }

    // If the alignment starts after the 3'-most breakpoint, then return 0/0/0/0
    if (read.getAlignmentStart() >= THREE_PRIME_EDGE_BP_LOC + 1) {
      return new BreakpointReads(0, 0, 0, 0);
    }

    // Whether or not the alignment starts before the 5'-most breakpoint
    boolean alignmentStartsBeforeFivePrimeBp = read.getAlignmentStart() <= FIVE_PRIME_EDGE_BP_LOC;

    // Initialize pos to alignment start position
    int pos = read.getAlignmentStart();

    // Boolean to keep track of if CIGAR has any clipping
    boolean isClipped = false;

    // Loop through CIGAR elements, updating isClipped boolean and the pos value
    for (CigarElement elt : read.getCigar().getCigarElements()) {
      switch (elt.getOperator()) {
        case H: // hard-clip: ignore
          isClipped = true;
          break;
        case S: // soft clip: return 0/1/0/0 if we are at the 3'-most breakpoint
          isClipped = true;
          if (pos == THREE_PRIME_EDGE_BP_LOC + 1) {
            return new BreakpointReads(0, 1, 0, 0);
          }
          break;
        case EQ: // sequence match: pass to alignment match
        case X: // sequence mismatch: pass to alignment match
        case M: // alignment match: add to pos
          pos += elt.getLength();
          break;
        case N: // reference skip: pass to deletion
        case D: // deletion: add to pos
          pos += elt.getLength();
          break;
        case P: // padding: don't add to pos
          break;
        case I: // insertion: don't add to pos
          break;
        default:
          throw new IllegalStateException("Case statement didn't deal with cigar op: " + elt.getOperator());
      }
    }

    // Whether or not the alignment spanned the 5'-most breakpoint
    boolean spansFivePrimeBp = alignmentStartsBeforeFivePrimeBp && pos > FIVE_PRIME_EDGE_BP_LOC + 1;

    // Whether or not the alignment spanned the 3'-most breakpoint
    boolean spansThreePrimeBp = pos > THREE_PRIME_EDGE_BP_LOC + 1;

    // If aligment spans neither breakpoint, then return 0/0/0/0
    if (!spansFivePrimeBp && !spansThreePrimeBp) {
      return new BreakpointReads(0, 0, 0, 0);

      // If alignment is clipped elsewhere and unclippedSpanningOnly is set, then return 0/0/0/0
    } else if (UNCLIPPED_SPANNING_ONLY && isClipped) {
      return new BreakpointReads(0, 0, 0, 0);

      // If alignment spans both breakpoints, the return 0/0/1/1
    } else if (spansFivePrimeBp && spansThreePrimeBp) {
      return new BreakpointReads(0, 0, 1, 1);

      // If alignment spans the 5'-most breakpoint only, the return 0/0/1/0
    } else if (spansFivePrimeBp) {
      return new BreakpointReads(0, 0, 1, 0);

      // If alignment spans the 3'-most breakpoint only, the return 0/0/0/1
    } else if (spansThreePrimeBp) {
      return new BreakpointReads(0, 0, 0, 1);

      // Should be unreachable code
    } else {
      throw new ShearRuntimeException("Unexpected behavior while calculating breakpoint reads");
    }

  }

  @Override
  public void onTraversalDone(BreakpointReads result) {

    // Output results to output file
    out.println(result.getFivePrimeClippedReads() + "\t" + result.getThreePrimeClippedReads() + "\t"
        + result.getFivePrimeSpanningReads() + "\t" + result.getThreePrimeSpanningReads());

    // Output results to standard output
    System.out.println(String.format(
        "Number of 5' end clipped reads at the breakpoint: %d",
        result.getFivePrimeClippedReads()));
    System.out.println(String.format(
        "Number of 3' end clipped reads at the breakpoint: %d",
        result.getThreePrimeClippedReads()));
    System.out.println(String.format(
        "Number of 5' end spanning reads at the breakpoint:    %d",
        result.getFivePrimeSpanningReads()));
    System.out.println(String.format(
        "Number of 3' end spanning reads at the breakpoint:    %d",
        result.getThreePrimeSpanningReads()));

  }

}
