/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.realignment;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Scanner;
import java.util.TreeSet;

import edu.umn.cs.kumarbio.Chromosome;
import edu.umn.cs.kumarbio.Genome;
import edu.umn.cs.kumarbio.shear.Shear;
import edu.umn.cs.kumarbio.shear.ShearRuntimeException;
import edu.umn.cs.kumarbio.shear.ShearSV;
import edu.umn.cs.kumarbio.shear.aligners.BwaAligner;
import edu.umn.cs.kumarbio.shear.aligners.BwaMemAligner;
import edu.umn.cs.kumarbio.shear.aligners.BwaSwAligner;
import edu.umn.cs.kumarbio.shear.heterogeneity.Breakpoint;
import edu.umn.cs.kumarbio.shear.system.RunGatkCommand;
import edu.umn.cs.kumarbio.shear.system.RunSystemCommand;
import edu.umn.cs.kumarbio.shear.variants.Variant;

/**
 * BreakpointsRealigner is used to extract relevant reads from a previous
 * alignment file and do local realignment on them to improve soft-clipping
 * accuracy and produce a new alignment file.
 */
public class BreakpointsRealigner {

  // Path to the BAM file to extract reads from
  private String                bamFilePath;

  // The Genome object to use
  private Genome                gen;

  // Prefix of the BWA index files to use
  private String                bwaIndexPrefix;

  // Alignment algorithm to use for local realignment
  private String                alignAlgo;

  // Whether or not to realign unmapped reads
  private boolean               realignUnmapped;

  // The instance of ShearSV that this is called from
  private ShearSV               shearSvInstance;

  // The number of bases to pad on either side of a breakpoint when extracting total reads
  private long                  mainExtractionPadding;

  // The number of bases to pad on either side of a breakpoint when extracting reads that are to be realigned
  private long                  realignmentExtractionPadding;

  // The number of bases that must be in between two adjacent realignment regions (otherwise they will be merged)
  private long                  realignmentRegionSeparation;

  // The number of bases to pad on either side of a realignment region when indexing the local region for the aligner
  private long                  realignmentRegionPadding;

  // The set of realignment regions
  private RealignmentRegionsSet regions;

  // The SAM header lines to use when necessary, taken from the input BAM file
  private ArrayList<String>     samHeaders;

  // Prefix to use for temporary files generated during extraction and realignment
  private String                tmpOutputPrefix;

  /**
   * Creates a new BreakpointsRealigner instance. Initialize
   * mainExtractionPadding to 3000, mainExtractionPadding to 10,
   * realignmentRegionSeparation to 5, and realignmentRegionPadding to 25.
   *
   * @param bamFilePath
   *          Path to the BAM file to extract reads from
   * @param gen
   *          The Genome object to use
   * @param picardDir
   *          The Picard Tools directory to use
   * @param bwaIndexPrefix
   *          Prefix of the BWA index files to use
   * @param alignAlgo
   *          Alignment algorithm to use for local realignment
   * @param shearSvInstance
   *          The instance of ShearSV that this is called from
   */
  public BreakpointsRealigner(
      String bamFilePath, Genome gen, String bwaIndexPrefix, String alignAlgo, boolean realignUnmapped,
      ShearSV shearSvInstance) {
    this.bamFilePath = bamFilePath;
    this.gen = gen;
    this.bwaIndexPrefix = bwaIndexPrefix;
    if (!(alignAlgo.equals("bwamem") || alignAlgo.equals("bwasw"))) {
      throw new ShearRuntimeException("Invalid alignment algorithm specified: " + alignAlgo);
    }
    this.alignAlgo = alignAlgo;
    this.realignUnmapped = realignUnmapped;
    this.shearSvInstance = shearSvInstance;
    mainExtractionPadding = 3000l;
    realignmentExtractionPadding = 10l;
    realignmentRegionSeparation = 5l;
    realignmentRegionPadding = 25l;
    regions = new RealignmentRegionsSet(gen, realignmentRegionSeparation);
    samHeaders = new ArrayList<String>();
    String rawSamHeaders = RunSystemCommand.run(new ProcessBuilder("samtools", "view", "-H", bamFilePath));
    String[] rawSamHeaderLines = rawSamHeaders.split("\n");
    for (String currentRawSamHeaderLine : rawSamHeaderLines) {
      if (currentRawSamHeaderLine.startsWith("@SQ") || currentRawSamHeaderLine.startsWith("@RG")) {
        samHeaders.add(currentRawSamHeaderLine);
      } else if (currentRawSamHeaderLine.startsWith("@HD")) {
        String tokens[] = currentRawSamHeaderLine.split("\t");
        String hdLine = "@HD";
        for (String token : tokens) {
          if (token.startsWith("VN")) {
            hdLine += "\t" + token;
          }
        }
        samHeaders.add(hdLine);
      }
    }
    samHeaders.add("@PG\tID:SHEAR\tPN:SHEAR\tVN:" + Shear.VERSION);
    tmpOutputPrefix = shearSvInstance.getTmpFilePrefix();
  }

  /**
   * Extracts reads around the breakpoints of the specified structural variants,
   * as well as unmapped reads (if realignUnmapped is set), realigns them and
   * outputs them to the specified SAM file. Reads from a large window around
   * the breakpoints are extracted, but the alignment for only those reads in a
   * small window around the breakpoints (plus the unmapped reads, if specified)
   * are realigned.
   *
   * @param svs
   *          The structural variants to process
   * @param outputSamFilePath
   *          The output file path for the resulting new alignment file in SAM
   *          format
   */
  public void extractAndRealign(TreeSet<Variant> svs, String outputSamFilePath) {

    // Get set of breakpoints from SVs
    TreeSet<Breakpoint> bps = new TreeSet<Breakpoint>();
    for (Variant sv : svs) {
      bps.addAll(sv.getBreakpoints());
    }

    // Ensure that there are breakpoints
    if (bps.size() == 0) {
      throw new ShearRuntimeException("Must be at least one breakpoint from SVs in order to extract and realign");
    }

    // Large BP windows SAM file
    String largeBpWindowsSamFile = tmpOutputPrefix + ".largeBpWindows.extracted.sam";
    prepareLargeBpWindowSamFile(bps, largeBpWindowsSamFile);

    // Realigned small BP windows SAM file
    String smallBpWindowsRealignedSamFile = tmpOutputPrefix + ".smallBpWindows.realigned.sam";
    prepareSmallBpWindowsRealignedwSamFile(bps, smallBpWindowsRealignedSamFile);

    // Realigned unmapped reads SAM file
    String unmappedRealignedSamFile = tmpOutputPrefix + ".unmapped.realigned.sam";
    if (realignUnmapped) {
      prepareUnmappedRealignedSamFile(unmappedRealignedSamFile);
    }

    // Store realigned breakpoint reads in a HashMap
    HashMap<String, SamRecord> bpRealignedReads = getAlignmentHashMap(smallBpWindowsRealignedSamFile);

    // Store realigned unmapped reads in a HashMap
    HashMap<String, SamRecord> unmappedRealignedReads = new HashMap<String, SamRecord>();
    if (realignUnmapped) {
      unmappedRealignedReads = getAlignmentHashMap(unmappedRealignedSamFile);
    }

    // Parse large BP windows SAM file, updating alignments for realigned reads, and outputting to final SAM file
    createNewAlignment(largeBpWindowsSamFile, bpRealignedReads, unmappedRealignedReads, outputSamFilePath);

    // Delete temporary files
    new File(largeBpWindowsSamFile).delete();
    new File(smallBpWindowsRealignedSamFile).delete();
    if (realignUnmapped) {
      new File(unmappedRealignedSamFile).delete();
    }

  }

  /**
   * Prepares a SAM file by extracting all relevant reads surrounding the
   * specified breakpoints using the mainExtractionPadding interval padding
   * (i.e. large breakpoint windows). Unmapped reads are not extracted.
   *
   * @param bps
   *          The set of breakpoints to extract from
   * @param outputSamFile
   *          The path to the desired SAM output file
   */
  private void prepareLargeBpWindowSamFile(TreeSet<Breakpoint> bps, String outputSamFile) {

    // Temporary files to use
    String tmpLargeBpWindowsExtractedBamFile = tmpOutputPrefix + ".largeBpWindows.extracted.bam";
    String tmpLargeBpWindowsExtractedBamIndexFile = tmpOutputPrefix + ".largeBpWindows.extracted.bai";

    // GATK log mode
    String gatkLogMode;
    if (shearSvInstance.isDebugMode()) {
      gatkLogMode = "INFO";
    } else {
      gatkLogMode = "ERROR";
    }

    try {

      // Extract all relevant reads (large breakpoint windows)
      // The --intervals argument is set to the range extending one base outside of the range of possible breakpoints
      // For example, if the break could be between 999 and 1000, or between 1000 and 1001, then --intervals is 999-1001
      // Note that the getFivePrimeLoc would return 999, and the getThreePrimeLoc would return 1000 (hence the +1 there)
      String gatkArgs =
          String.format(
              "-T PrintReads -I %s -R %s " + "--read_filter UnmappedRead --read_filter NotPrimaryAlignment "
                  + "--interval_padding %d -o %s -l %s",
              bamFilePath,
              gen.getFaFilePath(),
              mainExtractionPadding,
              tmpLargeBpWindowsExtractedBamFile,
              gatkLogMode);
      for (Breakpoint currentBp : bps) {
        gatkArgs +=
            String.format(
                " --intervals %s:%d-%d",
                currentBp.getChr().getName(),
                Math.max(1l, currentBp.getFivePrimeCoord()),
                Math.min(currentBp.getChr().getLength(), currentBp.getThreePrimeCoord() + 1));
      }
      RunGatkCommand.run(gatkArgs.split(" "), shearSvInstance.isDebugMode());

      // Write primary reads from the BAM file to the final SAM file
      FileOutputStream output = new FileOutputStream(outputSamFile);
      RunSystemCommand.run(
          new ProcessBuilder("samtools", "view", "-h", "-F", "3844", tmpLargeBpWindowsExtractedBamFile),
          output,
          "samtools",
          shearSvInstance.isDebugMode());
      output.close();

      // Catch errors
    } catch (FileNotFoundException e) {
      throw new ShearRuntimeException("Unexpected error during large breakpoint windows extraction", e);
    } catch (IOException e) {
      throw new ShearRuntimeException("Unexpected error during large breakpoint windows extraction", e);

      // Clean up
    } finally {

      // Clean up temporary files
      new File(tmpLargeBpWindowsExtractedBamFile).delete();
      new File(tmpLargeBpWindowsExtractedBamIndexFile).delete();

    }

  }

  /**
   * Prepares a SAM file by extracting all relevant reads surrounding the
   * specified breakpoints using the realignmentExtractionPadding interval
   * padding (i.e. small breakpoint windows), determining realignment regions,
   * realigning reads to their respective regions using a local realignment
   * algorithm, and finally merging the alignments and converting local region
   * coordinates to the proper global coordinates. Unmapped reads are not
   * extracted.
   *
   * @param bps
   *          The set of breakpoints to extract from
   * @param outputSamFile
   *          The path to the desired SAM output file
   */
  private void prepareSmallBpWindowsRealignedwSamFile(TreeSet<Breakpoint> bps, String outputSamFile) {

    // Temporary files to use
    String tmpSmallBpWindowsExtractedBamFile = tmpOutputPrefix + ".smallBpWindows.extracted.bam";
    String tmpSmallBpWindowsExtractedBamIndexFile = tmpOutputPrefix + ".smallBpWindows.extracted.bai";
    String tmpSmallBpWindowsExtractedSamFile = tmpOutputPrefix + ".smallBpWindows.extracted.sam";
    String tmpSmallBpWindowsAllRealignedSamFile = tmpOutputPrefix + ".smallBpWindows.all-realigned.sam";

    // GATK log mode
    String gatkLogMode;
    if (shearSvInstance.isDebugMode()) {
      gatkLogMode = "INFO";
    } else {
      gatkLogMode = "ERROR";
    }

    // Scanner object to use throughout
    Scanner sc = null;

    try {

      // Extract realignment reads (small breakpoint windows)
      // The --intervals argument is set to the range extending one base outside of the range of possible breakpoints
      // For example, if the break could be between 999 and 1000, or between 1000 and 1001, then --intervals is 999-1001
      // Note that the getFivePrimeLoc would return 999, and the getThreePrimeLoc would return 1000 (hence the +1 there)
      String gatkArgs =
          String.format(
              "-T PrintReads -I %s -R %s --read_filter UnmappedRead --read_filter NotPrimaryAlignment "
                  + "--interval_padding %d -o %s -l %s",
              bamFilePath,
              gen.getFaFilePath(),
              realignmentExtractionPadding,
              tmpSmallBpWindowsExtractedBamFile,
              gatkLogMode);
      for (Breakpoint currentBp : bps) {
        gatkArgs +=
            String.format(
                " --intervals %s:%d-%d",
                currentBp.getChr().getName(),
                Math.max(1l, currentBp.getFivePrimeCoord()),
                Math.min(currentBp.getChr().getLength(), currentBp.getThreePrimeCoord() + 1));
      }
      RunGatkCommand.run(gatkArgs.split(" "), shearSvInstance.isDebugMode());

      // Write primary reads from the BAM file to the SAM file
      FileOutputStream output = new FileOutputStream(tmpSmallBpWindowsExtractedSamFile);
      RunSystemCommand.run(
          new ProcessBuilder("samtools", "view", "-h", "-F", "3844", tmpSmallBpWindowsExtractedBamFile),
          output,
          "samtools",
          shearSvInstance.isDebugMode());
      output.close();

      // Determine realignment regions (first pass through small breakpoint windows SAM file)
      sc = new Scanner(new File(tmpSmallBpWindowsExtractedSamFile));
      while (sc.hasNextLine()) {
        String line = sc.nextLine();
        if (!line.startsWith("@")) {
          String[] samTokens = line.split("\t");
          regions.addRead(samTokens[2], Long.parseLong(samTokens[3]), samTokens[5]);
        }
      }
      sc.close();

      // All realignment regions
      ArrayList<RealignmentRegion> realignmentRegions = regions.getAllRealignmentRegions();

      // Open output streams for each realignment region FASTQ file
      HashMap<String, PrintWriter> realignmentReadsFastqOuts;
      realignmentReadsFastqOuts = new HashMap<String, PrintWriter>();
      for (RealignmentRegion currentRegion : realignmentRegions) {
        realignmentReadsFastqOuts.put(currentRegion.toString(), new PrintWriter(tmpOutputPrefix + ".smallBpWindows."
            + currentRegion.toString() + ".fq"));
      }

      // Convert realignment reads (SAM format) to multiple FASTQ format files for each realignment region
      // (second pass through small breakpoint windows SAM file)
      sc = new Scanner(new File(tmpSmallBpWindowsExtractedSamFile));
      while (sc.hasNextLine()) {
        String line = sc.nextLine();
        if (!line.startsWith("@")) {
          String[] samTokens = line.split("\t");
          RealignmentRegion realignmentRegion =
              regions.getRealignmentRegionForRead(samTokens[2], Long.parseLong(samTokens[3]));
          writeSamRecordToFastq(samTokens, realignmentReadsFastqOuts.get(realignmentRegion.toString()));
        }
      }
      sc.close();

      // Close output streams for each realignment region FASTQ file
      for (RealignmentRegion currentRegion : realignmentRegions) {
        realignmentReadsFastqOuts.get(currentRegion.toString()).close();
      }

      // Set up output stream for combined realigned reads SAM file
      PrintWriter outputWriter = new PrintWriter(tmpSmallBpWindowsAllRealignedSamFile);

      // Write SAM header lines to combined realigned reads SAM file
      for (String samHeaderLine : samHeaders) {
        outputWriter.println(samHeaderLine);
      }

      // Do local realignment for realignment regions
      for (RealignmentRegion currentRegion : realignmentRegions) {

        // Get realignment region information from the realignment region name
        Chromosome currentRegionChr = currentRegion.getStartLoc().getChr();
        long currentRegionStart = currentRegion.getStartLoc().getCoord();
        long currentRegionEnd = currentRegion.getEndLoc().getCoord();

        // Files to use for this region's local realignment
        String regionFastaFile = tmpOutputPrefix + ".smallBpWindows." + currentRegion.toString() + ".fa";
        String regionFastaIndexFile = tmpOutputPrefix + ".smallBpWindows." + currentRegion.toString() + ".fa.fai";
        String regionFastqFile = tmpOutputPrefix + ".smallBpWindows." + currentRegion.toString() + ".fq";
        String regionRealignedSamFile =
            tmpOutputPrefix + ".smallBpWindows." + currentRegion.toString() + ".realigned.sam";

        // Write region to FASTA file
        PrintWriter faOut = new PrintWriter(regionFastaFile);
        faOut.println(">" + currentRegionChr.getName());
        long paddedRegionStart = Math.max(1l, currentRegionStart - realignmentRegionPadding);
        long paddedRegionEnd = Math.min(currentRegionChr.getLength(), currentRegionEnd + realignmentRegionPadding);
        String regionSeq = currentRegionChr.getSequence(paddedRegionStart, paddedRegionEnd);
        faOut.println(regionSeq);
        faOut.close();

        // Perform local alignment for this region (low -T value because we know it can align in this region)
        BwaAligner aligner = null;
        if (alignAlgo.equals("bwamem")) {
          aligner = new BwaMemAligner(regionFastaFile, true, shearSvInstance);
        } else if (alignAlgo.equals("bwasw")) {
          aligner = new BwaSwAligner(regionFastaFile, true, shearSvInstance);
        }
        ArrayList<String> bwaArgs = new ArrayList<String>();
        if (alignAlgo.equals("bwamem")) {
          bwaArgs.add("-L");
          bwaArgs.add("2,2");
        } else if (alignAlgo.equals("bwasw")) {
          bwaArgs.add("-c");
          bwaArgs.add("2");
        }
        bwaArgs.add("-T");
        bwaArgs.add("10");
        aligner.align(regionFastqFile, regionRealignedSamFile, bwaArgs);
        aligner.close();

        // Merge with combined realigned reads SAM file (need to calculate position relative to full reference)
        sc = new Scanner(new File(regionRealignedSamFile));
        while (sc.hasNextLine()) {
          String line = sc.nextLine();
          if (!line.startsWith("@")) {
            String[] samTokens = line.split("\t");
            for (int k = 0; k < samTokens.length; k++) {
              if (k == 3) {
                long adjustedPos =
                    Long.valueOf(samTokens[k]) == 0l ? 0l : Long.valueOf(samTokens[k]) + paddedRegionStart - 1l;
                outputWriter.print(adjustedPos + "\t");
              } else if (k == samTokens.length - 1) {
                outputWriter.print(samTokens[k] + "\n");
              } else {
                outputWriter.print(samTokens[k] + "\t");
              }
            }
          }
        }
        sc.close();

        // Clean up temporary files
        new File(regionFastaFile).delete();
        new File(regionFastaIndexFile).delete();
        new File(regionFastqFile).delete();
        new File(regionRealignedSamFile).delete();

      }

      // Close output stream for combined realigned reads SAM file
      outputWriter.close();

      // Write primary reads from the combined realigned reads SAM file to the final SAM file
      FileOutputStream finalOutput = new FileOutputStream(outputSamFile);
      RunSystemCommand.run(new ProcessBuilder(
          "samtools",
          "view",
          "-h",
          "-S",
          "-F",
          "3844",
          tmpSmallBpWindowsAllRealignedSamFile), finalOutput, "samtools", shearSvInstance.isDebugMode());
      finalOutput.close();

      // Catch errors
    } catch (FileNotFoundException e) {
      throw new ShearRuntimeException("Unexpected error during breakpoint reads extraction and realignment", e);
    } catch (IOException e) {
      throw new ShearRuntimeException("Unexpected error during breakpoint reads extraction and realignment", e);

      // Clean up
    } finally {

      // Close scanner
      sc.close();

      // Clean up temporary files
      new File(tmpSmallBpWindowsExtractedBamFile).delete();
      new File(tmpSmallBpWindowsExtractedBamIndexFile).delete();
      new File(tmpSmallBpWindowsExtractedSamFile).delete();
      new File(tmpSmallBpWindowsAllRealignedSamFile).delete();

    }

  }

  /**
   * Prepares a SAM file by extracting all unmapped reads and then realigning
   * them using a local realignment algorithm. The returned SAM file will
   * contain only those reads which were mapped after the local realignment.
   *
   * @param outputSamFile
   *          The path to the desired SAM output file
   */
  private void prepareUnmappedRealignedSamFile(String outputSamFile) {

    // Temporary files to use
    String tmpUnmappedExtractedSamFile = tmpOutputPrefix + ".unmapped.extracted.sam";
    String tmpUnmappedExtractedFqFile = tmpOutputPrefix + ".unmapped.extracted.fq";
    String tmpUnmappedAllRealignedSamFile = tmpOutputPrefix + ".unmapped.all-realigned.sam";

    // Scanner object to use throughout
    Scanner sc = null;

    try {

      // Extract unmapped reads (SAM format)
      FileOutputStream tmpUnmappedExtractedSamOut = new FileOutputStream(tmpUnmappedExtractedSamFile);
      RunSystemCommand.run(
          new ProcessBuilder("samtools", "view", "-h", "-f", "4", bamFilePath),
          tmpUnmappedExtractedSamOut,
          "samtools",
          shearSvInstance.isDebugMode());
      tmpUnmappedExtractedSamOut.close();

      // Convert unmapped reads (SAM format) to unmapped reads (FASTQ format)
      sc = new Scanner(new File(tmpUnmappedExtractedSamFile));
      PrintWriter tmpUnmappedExtractedFqOut = new PrintWriter(tmpUnmappedExtractedFqFile);
      while (sc.hasNextLine()) {
        String line = sc.nextLine();
        if (!line.startsWith("@")) {
          String[] samTokens = line.split("\t");
          writeSamRecordToFastq(samTokens, tmpUnmappedExtractedFqOut);
        }
      }
      sc.close();
      tmpUnmappedExtractedFqOut.close();

      // Perform local realignment for unmapped reads
      BwaAligner unmappedReadAligner = null;
      if (alignAlgo.equals("bwamem")) {
        unmappedReadAligner = new BwaMemAligner(bwaIndexPrefix, false, shearSvInstance);
      } else if (alignAlgo.equals("bwasw")) {
        unmappedReadAligner = new BwaSwAligner(bwaIndexPrefix, false, shearSvInstance);
      }
      unmappedReadAligner.align(tmpUnmappedExtractedFqFile, tmpUnmappedAllRealignedSamFile);
      unmappedReadAligner.close();

      // Set up output streams for final SAM file (unmapped reads that were mapped after realignment)
      FileOutputStream output = new FileOutputStream(outputSamFile);
      PrintWriter outputWriter = new PrintWriter(output);

      // Write SAM header lines to final SAM file
      for (String samHeaderLine : samHeaders) {
        outputWriter.println(samHeaderLine);
      }
      outputWriter.flush();

      // Write mapped primary reads from the realignment to the final SAM file
      RunSystemCommand.run(
          new ProcessBuilder("samtools", "view", "-S", "-F", "3844", tmpUnmappedAllRealignedSamFile),
          output,
          "samtools",
          shearSvInstance.isDebugMode());

      // Close output streams for final SAM file
      outputWriter.close();
      output.close();

      // Catch errors
    } catch (FileNotFoundException e) {
      throw new ShearRuntimeException("Unexpected error during unmapped reads extraction and realignment", e);
    } catch (IOException e) {
      throw new ShearRuntimeException("Unexpected error during unmapped reads extraction and realignment", e);

      // Clean up
    } finally {

      // Close scanner
      sc.close();

      // Clean up temporary files
      new File(tmpUnmappedExtractedSamFile).delete();
      new File(tmpUnmappedExtractedFqFile).delete();
      new File(tmpUnmappedAllRealignedSamFile).delete();

    }

  }

  /**
   * Converts a SAM-format read to a FASTQ-format read and writes it to the
   * specified output stream. This will figure out what to name the ID for the
   * read (i.e. append a '-1' or '-2' flag) and handle reverse-complement SAM
   * records.
   *
   * @param samTokens
   *          The SAM record in a string array format (i.e. after splitting a
   *          SAM record line on tab characters)
   * @param fastqOut
   *          The output stream to write to, in the form of a PrintWriter object
   */
  private void writeSamRecordToFastq(String[] samTokens, PrintWriter fastqOut) {

    // Write read ID with '-1' or '-2' flag
    if (isFirstOfPair(Integer.valueOf(samTokens[1]))) {
      fastqOut.println("@" + samTokens[0] + "-1");
    } else {
      fastqOut.println("@" + samTokens[0] + "-2");
    }

    // Write reverse-complemented read
    if (BigInteger.valueOf(Long.valueOf(samTokens[1]).byteValue()).testBit(4)) {
      fastqOut.println(Genome.rc(samTokens[9]));
      fastqOut.println("+");
      fastqOut.println(Genome.reverse(samTokens[10]));

      // Write regular read
    } else {
      fastqOut.println(samTokens[9]);
      fastqOut.println("+");
      fastqOut.println(samTokens[10]);
    }

  }

  /**
   * Returns true if the given SAM flag integer value reflects that it is the
   * first read in the pair, false otherwise.
   *
   * @param samFlag
   *          The integer SAM flag to check
   * @return True if the given SAM flag integer value reflects that it is the
   *         first read in the pair, false otherwise
   */
  private boolean isFirstOfPair(int samFlag) {
    return BigInteger.valueOf(Long.valueOf(samFlag).byteValue()).testBit(6);
  }

  /**
   * Reads the contents of a SAM file and returns a HashMap mapping the IDs of
   * the SAM records to the information about their full record.
   *
   * @param samFile
   *          The SAM file to process
   * @return A HashMap containing all of the SAM records, keyed by their IDs
   */
  private HashMap<String, SamRecord> getAlignmentHashMap(String samFile) {

    // HashMap mapping read IDs to their SAM record
    HashMap<String, SamRecord> bpRealignedReads = new HashMap<String, SamRecord>();

    // Scanner object to use
    Scanner sc = null;

    try {

      // Loop through SAM file, adding records to HashMap
      sc = new Scanner(new File(samFile));
      while (sc.hasNextLine()) {
        String line = sc.nextLine();
        if (!line.startsWith("@")) {
          String[] samTokens = line.split("\t");
          String optionalFields = samTokens[11];
          for (int i = 12; i < samTokens.length; i++) {
            optionalFields += "\t" + samTokens[i];
          }
          bpRealignedReads.put(samTokens[0], new SamRecord(
              samTokens[0],
              Integer.valueOf(samTokens[1]),
              samTokens[2],
              Long.valueOf(samTokens[3]),
              Integer.valueOf(samTokens[4]),
              samTokens[5],
              samTokens[6],
              Long.valueOf(samTokens[7]),
              Integer.valueOf(samTokens[8]),
              samTokens[9],
              samTokens[10],
              optionalFields));
        }
      }

      // Catch errors
    } catch (FileNotFoundException e) {
      throw new ShearRuntimeException("Unexpected error while processing realigned reads", e);
    }

    // Close scanner
    finally {
      sc.close();
    }

    return bpRealignedReads;

  }

  /**
   * Parses an input SAM file, outputting records as they are to the output SAM
   * file or correcting the alignment for reads that are present in the
   * breakpoint realigned set. Also includes the unmapped reads that were mapped
   * after realignment. If realignUnmapped is set, the unmapped reads hash map
   * should be empty.
   *
   * @param inputSamFile
   *          The input SAM file to use
   * @param bpRealignedReads
   *          The HashMap containing SAM records from the breakpoint realigned
   *          reads, keyed via "READ_ID-1" or "READ_ID-2"
   * @param unmappedRealignedReads
   *          The HashMap containing SAM records from the unmapped realigned
   *          reads, keyed via "READ_ID-1" or "READ_ID-2"
   * @param outputSamFile
   *          The output SAM file
   */
  private void createNewAlignment(
      String inputSamFile, HashMap<String, SamRecord> bpRealignedReads,
      HashMap<String, SamRecord> unmappedRealignedReads, String outputSamFile) {

    // Input and output streams to use
    Scanner sc = null;
    PrintWriter output = null;

    try {

      // Set up input and output streams
      sc = new Scanner(new File(inputSamFile));
      output = new PrintWriter(outputSamFile);

      // Print headers to output SAM file
      for (String samHeaderLine : samHeaders) {
        output.println(samHeaderLine);
      }

      // Loop through records in input SAM file
      while (sc.hasNextLine()) {
        String line = sc.nextLine();
        if (!line.startsWith("@")) {
          String[] samTokens = line.split("\t");
          String queryName = samTokens[0];
          int samFlag = Integer.valueOf(samTokens[1]);
          String mateReferenceName = samTokens[6];
          long matePos = Long.valueOf(samTokens[7]);
          int templateLength = Integer.valueOf(samTokens[8]);

          // Get the modified read ID to look for in HashMap keys (i.e. with '-1' or '-2' appended)
          String modifiedReadIdToCheck = queryName;
          if (isFirstOfPair(samFlag)) {
            modifiedReadIdToCheck += "-1";
          } else {
            modifiedReadIdToCheck += "-2";
          }

          // Loop for a potential breakpoint realigned read match
          // Set read to be processed, or display warning if it is processed twice
          SamRecord bpReadMatch = null;
          if (bpRealignedReads.containsKey(modifiedReadIdToCheck)) {
            bpReadMatch = bpRealignedReads.get(modifiedReadIdToCheck);
            if (!bpRealignedReads.get(modifiedReadIdToCheck).readProcessed) {
              bpRealignedReads.get(modifiedReadIdToCheck).readProcessed = true;
            } else {
              Shear.log("WARNING: Matching breakpoint read found twice, which shouldn't happen: "
                  + modifiedReadIdToCheck, shearSvInstance.isDebugMode());
            }
          }

          // Look for a potential unmapped realigned read match
          SamRecord unmappedReadMatch = null;
          if (unmappedRealignedReads.containsKey(modifiedReadIdToCheck)) {
            unmappedReadMatch = unmappedRealignedReads.get(modifiedReadIdToCheck);
          }

          // If unmapped read match, print a warning (shouldn't happen)
          if (unmappedReadMatch != null) {
            Shear.log("WARNING: Matching breakpoint read found in unmapped realigned data, which shouldn't happen: "
                + queryName, shearSvInstance.isDebugMode());
          }

          // If breakpoint realigned read is not a match match, print SAM record as is
          if (bpReadMatch == null) {
            output.println(line);

            // If breakpoint realigned read is a match, then use it to update the alignment
          } else if (bpReadMatch != null) {
            int newSamFlag = bpReadMatch.samFlag;
            if (BigInteger.valueOf(Long.valueOf(samFlag).byteValue()).testBit(0)) {
              newSamFlag += 1;
            }
            if (BigInteger.valueOf(Long.valueOf(samFlag).byteValue()).testBit(1)) {
              newSamFlag += 2;
            }
            if (BigInteger.valueOf(Long.valueOf(samFlag).byteValue()).testBit(3)) {
              newSamFlag += 8;
            }
            if (BigInteger.valueOf(Long.valueOf(samFlag).byteValue()).testBit(5)) {
              newSamFlag += 32;
            }
            if (BigInteger.valueOf(Long.valueOf(samFlag).byteValue()).testBit(6)) {
              newSamFlag += 64;
            }
            if (BigInteger.valueOf(Long.valueOf(samFlag).byteValue()).testBit(7)) {
              newSamFlag += 128;
            }
            output.printf(
                "%s\t%d\t%s\t%d\t%d\t%s\t%s\t%d\t%d\t%s\t%s\t%s\n",
                queryName,
                newSamFlag,
                bpReadMatch.referenceName,
                bpReadMatch.pos,
                bpReadMatch.mapq,
                bpReadMatch.cigar,
                mateReferenceName,
                matePos,
                templateLength,
                bpReadMatch.sequence,
                bpReadMatch.quality,
                bpReadMatch.optionalFields);
          }

        }
      }

      // Ensure that all breakpoint realigned reads were processed
      for (SamRecord read : bpRealignedReads.values()) {
        if (!read.readProcessed) {
          Shear.log("WARNING: Breakpoint realigned read was not processed while preparing realignment, "
              + "which shouldn't happen: " + read.queryName, shearSvInstance.isDebugMode());
        }
      }

      // Loop through unmapped realigned reads and add to the output SAM file
      for (SamRecord read : unmappedRealignedReads.values()) {
        String newQueryName;
        int newSamFlag = read.samFlag;
        newSamFlag += 1;
        newSamFlag += 8;
        if (read.queryName.endsWith("-1")) {
          newQueryName = read.queryName.substring(0, read.queryName.length() - 2);
          newSamFlag += 64;
        } else if (read.queryName.endsWith("-2")) {
          newQueryName = read.queryName.substring(0, read.queryName.length() - 2);
          newSamFlag += 128;
        } else {
          throw new ShearRuntimeException("Unexpected read ID in unmapped reads realignment file: " + read.queryName);
        }
        output.printf(
            "%s\t%d\t%s\t%d\t%d\t%s\t%s\t%d\t%d\t%s\t%s\t%s\n",
            newQueryName,
            newSamFlag,
            read.referenceName,
            read.pos,
            read.mapq,
            read.cigar,
            read.mateReferenceName,
            read.matePos,
            read.templateLength,
            read.sequence,
            read.quality,
            read.optionalFields);
      }

      // Catch errors
    } catch (FileNotFoundException e) {
      throw new ShearRuntimeException("Unexpected error while extracting and realigning reads", e);

      // Clean up
    } finally {
      sc.close();
      output.close();
    }

  }

}
