/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.realignment;

import edu.umn.cs.kumarbio.shear.ShearRuntimeException;
import edu.umn.cs.kumarbio.shear.variants.Location;

/**
 * RealignmentRegion is used to keep track of realignment regions. This class
 * just contains the start and end locations for a particular realignment
 * region.
 */
public class RealignmentRegion {

  // The start location for this realignment region
  private Location startLoc;

  // The end location for this realignment region
  private Location endLoc;

  /**
   * Creates a new RealignmentRegion object.
   *
   * @param startCoord
   *          The start coordinate for this realignment region
   * @param endCoord
   *          The end coordinate for this realignment region
   */
  public RealignmentRegion(Location startLoc, Location endLoc) {

    // Ensure that start and end locations are on the same chromosome
    if (!startLoc.getChr().equals(endLoc.getChr())) {
      throw new ShearRuntimeException("Start and end location for a realignment region must be on the same chromosome");
    }

    // Set instance variables
    this.startLoc = startLoc;
    this.endLoc = endLoc;

  }

  /**
   * Returns the start location for this realignment region.
   *
   * @return The start location for this realignment region
   */
  public Location getStartLoc() {
    return startLoc;
  }

  /**
   * Returns the end location for this realignment region.
   *
   * @return The end location for this realignment region
   */
  public Location getEndLoc() {
    return endLoc;
  }

  /**
   * Updates the coordinate of the start location for this realignment region,
   * ensuring that the new coordinate occurs within the bounds of the
   * chromosome.
   *
   * @param newStartCoord
   *          The coordinate of the new start location
   */
  public void updateStartCoord(long newStartCoord) {
    if (newStartCoord > getStartLoc().getChr().getLength()) {
      startLoc = new Location(getStartLoc().getChr(), getStartLoc().getChr().getLength());
    } else if (newStartCoord <= 0) {
      startLoc = new Location(getStartLoc().getChr(), 1);
    } else {
      startLoc = new Location(getStartLoc().getChr(), newStartCoord);
    }
  }

  /**
   * Updates the coordinate of the end location for this realignment region,
   * ensuring that the new coordinate occurs within the bounds of the
   * chromosome.
   *
   * @param newEndCoord
   *          The coordinate of the new end location
   */
  public void updateEndCoord(long newEndCoord) {
    if (newEndCoord > getEndLoc().getChr().getLength()) {
      endLoc = new Location(getEndLoc().getChr(), getEndLoc().getChr().getLength());
    } else if (newEndCoord <= 0) {
      endLoc = new Location(getEndLoc().getChr(), 1);
    } else {
      endLoc = new Location(getEndLoc().getChr(), newEndCoord);
    }
  }

  @Override
  public String toString() {
    return getStartLoc().getChr().getName() + "_" + getStartLoc().getCoord() + "_" + getEndLoc().getCoord();
  }

}
