/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.system;

import java.io.IOException;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.io.PrintStream;
import java.util.concurrent.FutureTask;

import org.broadinstitute.gatk.utils.commandline.CommandLineProgram;

import edu.umn.cs.kumarbio.shear.Shear;
import edu.umn.cs.kumarbio.shear.ShearRuntimeException;

/**
 * RunGatkCommand is used to run a GATK command. It contains only the static run
 * method, which takes a string array specifying the GATK parameters and runs
 * it, redirecting the output and error streams to be prepended with a tag and
 * sent to standard output.
 */
public class RunGatkCommand {

  /**
   * Runs the designated GATK command.
   *
   * @param gatkArgs
   *          Array of arguments for the GATK command
   * @param debug
   *          Whether or not debugging mode is on
   */
  public static void run(String[] gatkArgs, boolean debug) {

    // Print out the command being executed
    if (debug) {
      StringBuilder command = new StringBuilder("  [DEBUG] GATK-COMMAND:");
      for (int i = 0; i < gatkArgs.length; i++) {
        command.append(" " + gatkArgs[i]);
      }
      Shear.log(command.toString(), false);
    }

    // Save original standard out/err streams
    PrintStream origSystemOut = System.out;
    PrintStream origSystemErr = System.err;

    // Set up piped output->input streams
    PipedOutputStream pipedOut = new PipedOutputStream();
    PipedInputStream pipedIn = new PipedInputStream();

    // Wrap piped output in a PrintStream object
    PrintStream printStream = new PrintStream(pipedOut);

    // Redirector in separate thread
    FutureTask<Boolean> redirector = null;

    try {

      // Connect piped output->input streams
      pipedOut.connect(pipedIn);

      // Point tagged out/err to piped output
      System.setOut(printStream);
      System.setErr(printStream);

      // Redirect the output stream in separate thread
      PrintStream redirectionTarget = debug ? origSystemOut : null;
      redirector = new FutureTask<Boolean>(new StreamRedirector(pipedIn, redirectionTarget, "GATK"));
      new Thread(redirector).start();

      // Run GATK command
      CustomCommandLineGATK gatk = new CustomCommandLineGATK();
      CommandLineProgram.start(gatk, gatkArgs);
      gatk.closeBamFiles();

      // Catch IO errors
    } catch (IOException e) {
      throw new ShearRuntimeException(
          "Error encountered with PipedOutputStream and/or PipedInputStream objects for stream redirection",
          e);
    } catch (Exception e) {
      throw new ShearRuntimeException("Error with running GATK command", e);

      // Clean up
    } finally {

      // Close tagged out/err streams
      System.out.close();
      System.err.close();

      // Wait for the streams
      boolean waitForStreamFailed = false;
      Exception waitForStreamException = null;
      try {
        redirector.get();
      } catch (Exception e) {
        waitForStreamFailed = true;
        waitForStreamException = e;
      }

      // Close piped output->input streams
      printStream.close();

      // Reset original standard out/err streams
      System.setOut(origSystemOut);
      System.setErr(origSystemErr);

      // Print ending blank line
      if (debug) {
        System.out.println("[SHEAR]");
      }

      // Throw error if waiting for streams failed
      if (waitForStreamFailed) {
        throw new ShearRuntimeException("Error while waiting for stream redirection", waitForStreamException);
      }

    }

  }
}
