/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.variantcallers;

import htsjdk.samtools.CigarElement;
import htsjdk.samtools.CigarOperator;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Scanner;
import java.util.TreeSet;

import org.broadinstitute.gatk.utils.smithwaterman.SWPairwiseAlignment;

import edu.umn.cs.kumarbio.Genome;
import edu.umn.cs.kumarbio.shear.Shear;
import edu.umn.cs.kumarbio.shear.ShearRuntimeException;
import edu.umn.cs.kumarbio.shear.system.RunSystemCommand;
import edu.umn.cs.kumarbio.shear.variants.Deletion;
import edu.umn.cs.kumarbio.shear.variants.IllegalVariantException;
import edu.umn.cs.kumarbio.shear.variants.Inversion;
import edu.umn.cs.kumarbio.shear.variants.Location;
import edu.umn.cs.kumarbio.shear.variants.TandemDuplication;
import edu.umn.cs.kumarbio.shear.variants.Translocation;
import edu.umn.cs.kumarbio.shear.variants.Variant;

/**
 * This is a wrapper for the CREST SV caller. This class extends the abstract
 * class VariantCaller.
 */
public class CrestWrapper extends VariantCaller {

  // The Genome object to use for this instance of CREST
  private Genome   gen;

  // Whether or not debugging mode is active
  private boolean  debug;

  // The average read length for the data to process
  private int      readLength;

  // The directory that contains the CREST scripts to use
  private String   crestDir;

  // The prefix to use for log and output files for the GFServer
  private String   gfServerPrefix;

  // The *.2bit file used by the GFServer
  private String   gfServerTwoBitFile;

  // The GFServer object to use for this instance of CREST
  private GFServer gfServer;

  /**
   * Creates a new CrestWrapper using the specified genome. The FASTA reference
   * file associated the genome is required to have an accompanying *.fai index
   * file. Also requires the directory that contains the CREST scripts, as well
   * as the average read length of the data to process and information required
   * to start up the GFServer.
   *
   * @param gen
   *          The Genome object to use for this instance of CREST
   * @param debug
   *          Whether or not debugging mode is active
   * @param readLength
   *          The average read length for the data to process
   * @param crestDir
   *          The directory that contains the CREST scripts to use
   * @param gfServerPrefix
   *          The prefix to use for log and output files for the GFServer
   * @param gfServerTwoBitFile
   *          The *.2bit file used by the GFServer
   */
  public CrestWrapper(
      Genome gen, boolean debug, int readLength, String crestDir, String gfServerPrefix, String gfServerTwoBitFile) {

    // Set instance variables
    this.gen = gen;
    this.debug = debug;
    this.readLength = readLength;
    this.crestDir = crestDir;
    this.gfServerPrefix = gfServerPrefix;
    this.gfServerTwoBitFile = gfServerTwoBitFile;

    // Initialize GFServer
    gfServer = initializeGFServer();

  }

  /**
   * Sets up an instance of GFServer to be used for this instance of CREST.
   *
   * @return The GFServer object to use for this instance of CREST after it is
   *         set up
   */
  private GFServer initializeGFServer() {

    try {

      // Initialize GFServer object
      GFServer gfServer = new GFServer(gfServerPrefix, gfServerTwoBitFile);

      // Start GFServer
      if (debug) {
        Shear.log("[DEBUG] Using port " + gfServer.getPortNum() + " for gfServer", false);
        Shear.log("[DEBUG] Starting gfServer...", false);
      }
      gfServer.start();

      // Wait for GFServer to be ready
      if (debug) {
        Shear.log("[DEBUG] Waiting for gfServer to be ready...", false);
      }
      boolean gfServerStarted = gfServer.waitForReady();

      // Loop for every GFServer that was not successfully started
      while (!gfServerStarted) {

        // Throw error if *.2bit file is invalid
        if (!gfServer.isValidTwoBitFile()) {
          throw new ShearRuntimeException("Not a valid two-bit reference file: " + gfServerTwoBitFile);
        }

        // If bad port number was used, clear log files and close GFServer
        if (debug) {
          Shear.log("[DEBUG] Port already in use. Restarting gfServer...", false);
        }
        gfServer.clearLogFiles();
        gfServer.close();

        // Initialize new GFServer object starting at the next port number after the failed one
        gfServer = new GFServer(gfServerPrefix, gfServerTwoBitFile, gfServer.getPortNum() + 1);

        // Start GFServer
        if (debug) {
          Shear.log("[DEBUG] Using port " + gfServer.getPortNum() + " for gfServer", false);
          Shear.log("[DEBUG] Starting gfServer...", false);
        }
        gfServer.start();

        // Wait for new GFServer to be ready
        if (debug) {
          Shear.log("[DEBUG] Waiting for gfServer to be ready...", false);
        }
        gfServerStarted = gfServer.waitForReady();

      }

      // Print debugging log message that GFServer is ready
      if (debug) {
        Shear.log("[DEBUG] gfServer is ready!", false);
      }

      // Return GFServer object
      return gfServer;

      // Catch possible errors
    } catch (FileNotFoundException e) {
      throw new ShearRuntimeException("Unable to create GFServer to be used by CREST", e);
    } catch (InterruptedException e) {
      throw new ShearRuntimeException("Unable to create GFServer to be used by CREST", e);
    }

  }

  /**
   * Detects microhomology for an SV and returns the number of homologous bases.
   * This works by checking the number of bases on either side of the breakpoint
   * that the consensus sequence aligns with and comparing it against the length
   * of the consensus sequence. If the length of the consensus sequence is
   * shorter than the sum of the aligned bases on either side of the breakpoint
   * then that means that part of the consensus sequence is aligned twice. These
   * are the homologous bases. This will return 0 if anything unexpected
   * happens.
   *
   * @param chr1
   *          The name of the chromosome for the first breakpoint location as
   *          reported by CREST
   * @param loc1
   *          The first breakpoint location as reported by CREST
   * @param strand1
   *          The strand of the first breakpoint as reported by CREST
   * @param chr2
   *          The name of the chromosome for the second breakpoint location as
   *          reported by CREST
   * @param loc2
   *          The second breakpoint location as reported by CREST
   * @param strand2
   *          The strand of the second breakpoint as reported by CREST
   * @param consensusSeq
   *          The consensus sequence for this variant as reported by CREST
   * @return The number of homologous bases at the SV breakpoints
   */
  private int detectMicrohomology(
      String chr1, long loc1, String strand1, String chr2, long loc2, String strand2, String consensusSeq) {

    try {

      // Objects to hold the alignments
      SWPairwiseAlignment sw1 = null;
      SWPairwiseAlignment sw2 = null;

      // If both strands are '+'
      if (strand1.equals("+") && strand2.equals("+")) {

        // Get the bases that border the breakpoints on the non-clipped side
        String bp1bases = gen.getSequence(chr1, loc1 - readLength + 1, loc1);
        String bp2bases = gen.getSequence(chr2, loc2, loc2 + readLength - 1);

        // Get alignments of the breakpoint reads against the consensus sequence
        // bp1bases should align at first half of consensusSeq (thus reverse both to easily get the middle loc)
        // bp2bases should align at second half of consensusSeq
        sw1 = new SWPairwiseAlignment(Genome.reverse(consensusSeq).getBytes(), Genome.reverse(bp1bases).getBytes());
        sw2 = new SWPairwiseAlignment(consensusSeq.getBytes(), bp2bases.getBytes());

        // If strand1 is '+' and strand2 is '-'
      } else if (strand1.equals("+") && strand2.equals("-")) {

        // Get the bases that border the breakpoints on the non-clipped side
        String bp1bases = gen.getSequence(chr1, loc1 - readLength + 1, loc1);
        String bp2bases = gen.getSequence(chr2, loc2 - readLength + 1, loc2);

        // Assume that consensus sequence starts with first breakpoint forward strand reads
        // bp1bases should align at first half of consensusSeq (thus reverse both to easily get the middle loc)
        // Reverse-complement of bp2bases should align at second half of consensusSeq
        SWPairwiseAlignment attempt1sw1 =
            new SWPairwiseAlignment(Genome.reverse(consensusSeq).getBytes(), Genome.reverse(bp1bases).getBytes());
        SWPairwiseAlignment attempt1sw2 =
            new SWPairwiseAlignment(consensusSeq.getBytes(), Genome.rc(bp2bases).getBytes());

        // Assume that consensus sequence starts with second breakpoint forward strand reads
        // Reverse-complement of bp1bases should align at second half of consensusSeq
        // bp2bases should align at first half of consensusSeq (thus reverse both to easily get the middle loc)
        SWPairwiseAlignment attempt2sw1 =
            new SWPairwiseAlignment(consensusSeq.getBytes(), Genome.rc(bp1bases).getBytes());
        SWPairwiseAlignment attempt2sw2 =
            new SWPairwiseAlignment(Genome.reverse(consensusSeq).getBytes(), Genome.reverse(bp2bases).getBytes());

        // Painful hack, but just choose the alignments that result in shorter CIGAR strings overall
        // Long CIGAR strings indicate a poor match and that it is the wrong assumption about the consensus sequence
        int attempt1cigarStringLength =
            attempt1sw1.getCigar().toString().length() + attempt1sw2.getCigar().toString().length();
        int attempt2cigarStringLength =
            attempt2sw1.getCigar().toString().length() + attempt2sw2.getCigar().toString().length();
        if (attempt1cigarStringLength <= attempt2cigarStringLength) {
          sw1 = attempt1sw1;
          sw2 = attempt1sw2;
        } else {
          sw1 = attempt2sw1;
          sw2 = attempt2sw2;
        }

        // If strand1 is '-' and strand2 is '+'
      } else if (strand1.equals("-") && strand2.equals("+")) {

        // Get the bases that border the breakpoints on the non-clipped side
        String bp1bases = gen.getSequence(chr1, loc1, loc1 + readLength - 1);
        String bp2bases = gen.getSequence(chr2, loc2, loc2 + readLength - 1);

        // Assume that consensus sequence starts with first breakpoint reverse strand reads
        // Reverse-complement of bp1bases should align at first half of consensusSeq (thus reverse both ref, comp bases)
        // bp2bases should align at second half of consensusSeq
        SWPairwiseAlignment attempt1sw1 =
            new SWPairwiseAlignment(Genome.reverse(consensusSeq).getBytes(), Genome.complement(bp1bases).getBytes());
        SWPairwiseAlignment attempt1sw2 = new SWPairwiseAlignment(consensusSeq.getBytes(), bp2bases.getBytes());

        // Assume that consensus sequence starts with second breakpoint forward strand reads
        // bp1bases should align at second half of consensusSeq
        // Reverse-complement of bp2bases should align at first half of consensusSeq (thus reverse both ref, comp bases)
        SWPairwiseAlignment attempt2sw1 = new SWPairwiseAlignment(consensusSeq.getBytes(), bp1bases.getBytes());
        SWPairwiseAlignment attempt2sw2 =
            new SWPairwiseAlignment(Genome.reverse(consensusSeq).getBytes(), Genome.complement(bp2bases).getBytes());

        // Painful hack, but just choose the alignments that result in shorter CIGAR strings overall
        // Long CIGAR strings indicate a poor match and that it is the wrong assumption about the consensus sequence
        int attempt1cigarStringLength =
            attempt1sw1.getCigar().toString().length() + attempt1sw2.getCigar().toString().length();
        int attempt2cigarStringLength =
            attempt2sw1.getCigar().toString().length() + attempt2sw2.getCigar().toString().length();
        if (attempt1cigarStringLength <= attempt2cigarStringLength) {
          sw1 = attempt1sw1;
          sw2 = attempt1sw2;
        } else {
          sw1 = attempt2sw1;
          sw2 = attempt2sw2;
        }

        // If strands are invalid
      } else {
        return 0;
      }

      // Get the number of aligned bases
      // Aligned bases are the number of consensus sequence bases that are aligned with breakpoint bases (measured both ends)
      // If aligned bases using both ends are greater than the consensus sequence length, then there is microhomology
      int alignedBases1 = consensusSeq.length() - sw1.getAlignmentStart2wrt1();
      int alignedBases2 = consensusSeq.length() - sw2.getAlignmentStart2wrt1();

      // Validate that alignments make sense
      // If aligned bases using both ends are greater than the consensus sequence length, then there is microhomology
      CigarElement cigarEl1 = sw1.getCigar().getCigarElement(0);
      CigarElement cigarEl2 = sw2.getCigar().getCigarElement(0);
      if (alignedBases1 <= 0 || alignedBases2 <= 0 || cigarEl1.getOperator() != CigarOperator.MATCH_OR_MISMATCH
          || cigarEl1.getLength() < 5 || cigarEl2.getOperator() != CigarOperator.MATCH_OR_MISMATCH
          || cigarEl2.getLength() < 5 || consensusSeq.length() > alignedBases1 + alignedBases2) {
        return 0;
      }

      // Return
      return alignedBases1 + alignedBases2 - consensusSeq.length();

      // Catch error
    } catch (IOException e) {
      throw new ShearRuntimeException("Error occurred while reading from the reference sequence file", e);
    }

  }

  @Override
  public void close() {

    // Display log message
    if (debug) {
      Shear.log("[DEBUG] Closing gfServer...", false);
    }

    // Close GFServer
    try {
      gfServer.close();
    } catch (Exception e) {
      throw new ShearRuntimeException("Error while trying to close down GFServer");
    }

    // Delete log files
    if (!debug) {
      gfServer.clearLogFiles();
    }

  }

  @Override
  public String runVariantCaller(String bamFile, String outputPrefix, List<String> options) {

    // Output files to use
    String crestSclipFilePath = outputPrefix + ".sclip.txt";
    String crestCoverFilePath = outputPrefix + ".cover";
    String crestPredFilePath = outputPrefix + ".predSV.txt";

    // Delete files that may be present from a previous run
    new File(crestSclipFilePath).delete();
    new File(crestCoverFilePath).delete();
    new File(crestPredFilePath).delete();

    // Run CREST's extractSClip.pl script
    ProcessBuilder extractSClipProcess =
        new ProcessBuilder(
            "perl",
            "-I" + crestDir,
            crestDir + "/extractSClip.pl",
            "-p",
            outputPrefix,
            "-i",
            bamFile,
            "--ref_genome",
            gen.getFaFilePath());
    RunSystemCommand.run(extractSClipProcess, System.out, "CREST", debug);

    // Prepare list of command and arguments for CREST.pl script
    ArrayList<String> crestCommandAndArgs = new ArrayList<String>();
    crestCommandAndArgs.add("perl");
    crestCommandAndArgs.add("-I" + crestDir);
    crestCommandAndArgs.add(crestDir + "/CREST.pl");
    crestCommandAndArgs.add("--read_len");
    crestCommandAndArgs.add(String.valueOf(readLength));
    crestCommandAndArgs.add("--max_rep_cover");
    crestCommandAndArgs.add("1000000");
    for (int i = 0; i < options.size(); i++) {
      crestCommandAndArgs.add(options.get(i));
    }
    crestCommandAndArgs.add("-p");
    crestCommandAndArgs.add(outputPrefix);
    crestCommandAndArgs.add("-f");
    crestCommandAndArgs.add(crestCoverFilePath);
    crestCommandAndArgs.add("-d");
    crestCommandAndArgs.add(bamFile);
    crestCommandAndArgs.add("--ref_genome");
    crestCommandAndArgs.add(gen.getFaFilePath());
    crestCommandAndArgs.add("-t");
    crestCommandAndArgs.add(gfServer.getTwoBitFile());
    crestCommandAndArgs.add("--blatserver");
    crestCommandAndArgs.add("localhost");
    crestCommandAndArgs.add("--blatport");
    crestCommandAndArgs.add(String.valueOf(gfServer.getPortNum()));

    // Run CREST's CREST.pl script
    ProcessBuilder crestProcess = new ProcessBuilder(crestCommandAndArgs);
    RunSystemCommand.run(crestProcess, System.out, "CREST", debug);

    // Clean up extractSClip.pl output files if not in debug mode
    if (!debug) {
      new File(crestSclipFilePath).delete();
      new File(crestCoverFilePath).delete();
    }

    // Return output file path
    return crestPredFilePath;

  }

  @Override
  public TreeSet<Variant> getVariants(String variantCallerOutputFile) {

    // Note that this getVariants implementation DOES NOT calculate heterogeneity percentages for variants

    // Initialize set of variants to return
    TreeSet<Variant> variants = new TreeSet<Variant>();

    // Open up a scanner for predSV.txt output file and read from it
    Scanner sc = null;
    try {
      sc = new Scanner(new File(variantCallerOutputFile));
      while (sc.hasNextLine()) {
        String line = sc.nextLine();
        String[] tokens = line.split("\t");

        // Validate that line has expected format
        if (tokens.length >= 24 && tokens[1].matches("^[0-9]+$") && tokens[5].matches("^[0-9]+$")
            && tokens[2].matches("^\\+|-$") && tokens[6].matches("^\\+|-$") && tokens[23].matches("^[ACGTNXacgtnx]+$")) {

          // Get relevant predSV.txt line information
          String chr1 = tokens[0];
          long loc1 = Long.valueOf(tokens[1]);
          String strand1 = tokens[2];
          String chr2 = tokens[4];
          long loc2 = Long.valueOf(tokens[5]);
          String strand2 = tokens[6];
          String svType = tokens[8];
          String consensusSeq = tokens[23];

          // The variant object
          Variant v = null;

          try {

            // Create Deletion
            if (svType.equals("DEL") && chr1.equals(chr2) && loc1 < loc2 - 1 && strand1.equals("+")
                && strand2.equals("+")) {
              long bpAdjustment = detectMicrohomology(chr1, loc1, strand1, chr2, loc2, strand2, consensusSeq);
              v =
                  new Deletion(new Location(gen.getChromosome(chr1), loc1 + 1), loc2 - loc1 - 1 + bpAdjustment, false)
                      .getCanonicalEquivalent();

              // Create Tandem Duplication
            } else if (svType.equals("INS") && chr1.equals(chr2) && loc1 >= loc2 && strand1.equals("+")
                && strand2.equals("+")) {
              long bpAdjustment = detectMicrohomology(chr1, loc1, strand1, chr2, loc2, strand2, consensusSeq);
              v =
                  new TandemDuplication(new Location(gen.getChromosome(chr2), loc2), loc1 - loc2 + 1 - bpAdjustment)
                      .getCanonicalEquivalent();

              // Create Inversion (type 1, + > -)
            } else if (svType.equals("INV") && chr1.equals(chr2) && loc1 < loc2 - 1 && strand1.equals("+")
                && strand2.equals("-")) {
              v =
                  new Inversion(new Location(gen.getChromosome(chr1), loc1 + 1), loc2 - loc1 - 1)
                      .getCanonicalEquivalent();

              // Create Inversion (type 2, - > +)
            } else if (svType.equals("INV") && chr1.equals(chr2) && loc1 >= loc2 && strand1.equals("-")
                && strand2.equals("+")) {
              v = new Inversion(new Location(gen.getChromosome(chr2), loc2), loc1 - loc2 + 1).getCanonicalEquivalent();

              // Create Intra-Chromosome Translocation (type 1, + > -)
            } else if (svType.equals("ITX") && chr1.equals(chr2) && strand1.equals("+") && strand2.equals("-")) {
              long bpAdjustment = detectMicrohomology(chr1, loc1, strand1, chr2, loc2, strand2, consensusSeq);
              v =
                  new Translocation(new Location(gen.getChromosome(chr1), loc1 - bpAdjustment), new Location(
                      gen.getChromosome(chr2),
                      loc2), Translocation.Type.for3toRev5).getCanonicalEquivalent();

              // Create Intra-Chromosome Translocation (type 2, - > +)
            } else if (svType.equals("ITX") && chr1.equals(chr2) && strand1.equals("-") && strand2.equals("+")) {
              long bpAdjustment = detectMicrohomology(chr1, loc1, strand1, chr2, loc2, strand2, consensusSeq);
              v =
                  new Translocation(new Location(gen.getChromosome(chr2), loc2 + bpAdjustment), new Location(
                      gen.getChromosome(chr1),
                      loc1), Translocation.Type.for5toRev3).getCanonicalEquivalent();

              // Create Inter-Chromosome Translocation (type 1, + > -)
            } else if (svType.equals("CTX") && !chr1.equals(chr2) && strand1.equals("+") && strand2.equals("-")) {
              long bpAdjustment = detectMicrohomology(chr1, loc1, strand1, chr2, loc2, strand2, consensusSeq);
              v =
                  new Translocation(new Location(gen.getChromosome(chr1), loc1 - bpAdjustment), new Location(
                      gen.getChromosome(chr2),
                      loc2), Translocation.Type.for3toRev5).getCanonicalEquivalent();

              // Create Inter-Chromosome Translocation (type 2, - > +)
            } else if (svType.equals("CTX") && !chr1.equals(chr2) && strand1.equals("-") && strand2.equals("+")) {
              long bpAdjustment = detectMicrohomology(chr1, loc1, strand1, chr2, loc2, strand2, consensusSeq);
              v =
                  new Translocation(new Location(gen.getChromosome(chr2), loc2 + bpAdjustment), new Location(
                      gen.getChromosome(chr1),
                      loc1), Translocation.Type.for5toRev3).getCanonicalEquivalent();

              // Create Inter-Chromosome Translocation (type 3, + > +)
            } else if (svType.equals("CTX") && !chr1.equals(chr2) && strand1.equals("+") && strand2.equals("+")) {
              long bpAdjustment = detectMicrohomology(chr1, loc1, strand1, chr2, loc2, strand2, consensusSeq);
              v =
                  new Translocation(new Location(gen.getChromosome(chr1), loc1 - bpAdjustment), new Location(
                      gen.getChromosome(chr2),
                      loc2), Translocation.Type.for3toFor5).getCanonicalEquivalent();

              // Skip over lines with unexpected format
            } else {
              Shear.log("WARNING: Skipping unexpected event in CREST output (" + tokens[0] + "\t" + tokens[1] + "\t"
                  + tokens[2] + "\t" + tokens[3] + "\t" + tokens[4] + "\t" + tokens[5] + "\t" + tokens[6] + "\t"
                  + tokens[7] + "\t" + tokens[8] + ")", debug);
              continue;

            }

            // Skip over line that caused IllegalVariantException error
          } catch (IllegalVariantException e) {
            Shear.log("[DEBUG] WARNING: Skipping illegal variant in CREST output (" + tokens[0] + "\t" + tokens[1]
                + "\t" + tokens[2] + "\t" + tokens[3] + "\t" + tokens[4] + "\t" + tokens[5] + "\t" + tokens[6] + "\t"
                + tokens[7] + "\t" + tokens[8] + ")", debug);
            continue;

          }

          // Add variant to return set
          variants.add(v);

          // Throw error for unexpected predSV.txt format
        } else {
          throw new ShearRuntimeException("Unexpected format in predSV.txt output file '" + variantCallerOutputFile
              + "' from CREST");
        }

      }

      // Catch error from setting up Scanner object for predSV.txt output
    } catch (FileNotFoundException e) {
      throw new ShearRuntimeException("Unable to locate predSV.txt output file '" + variantCallerOutputFile
          + "' from CREST", e);

    } finally {

      // Close Scanner object
      if (sc != null) {
        sc.close();
      }

    }

    // Return variants
    return variants;

  }

}
