/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.variantcallers;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.util.Scanner;

import edu.umn.cs.kumarbio.shear.ShearRuntimeException;
import edu.umn.cs.kumarbio.shear.system.RunSystemCommand;

/**
 * GFServer is used to run the gfserver process in a separate thread. This is
 * used by CREST.
 */
public class GFServer extends Thread {

  // The prefix to use for log and output files for this GFServer
  private final String         prefix;

  // The *.2bit file used by this GFServer
  private final String         twoBitFile;

  // Whether or not the *.2bit file used by this GFServer is valid
  private boolean              validTwoBitFile;

  // The port number used by this GFServer
  private int                  portNum;

  // The ProcessBuilder object to run this GFserver with
  private final ProcessBuilder pb;

  // The Process object that runs this GFServer
  private Process              proc;

  /**
   * Sets up a new GFServer object using the specified *.2bit file. Output and
   * log files are created based on the prefix string specified. This uses 49152
   * as the first port number to check when finding a port for this GFServer.
   *
   * @param prefix
   *          The prefix to use for log and output files for this GFServer
   * @param twoBitFile
   *          The *.2bit file used by this GFServer
   */
  public GFServer(String prefix, String twoBitFile) {
    this(prefix, twoBitFile, 49152);
  }

  /**
   * Sets up a new GFServer object using the specified *.2bit file. Output and
   * log files are created based on the prefix string specified.
   *
   * @param prefix
   *          The prefix to use for log and output files for this GFServer
   * @param twoBitFile
   *          The *.2bit file used by this GFServer
   * @param debug
   *          Whether or not debugging mode is active
   * @param startingPortNum
   *          The first port number to check when finding a port for this
   *          GFServer
   */
  public GFServer(String prefix, String twoBitFile, int startingPortNum) {

    // Set instance variables
    this.prefix = prefix;
    this.twoBitFile = twoBitFile;
    validTwoBitFile = true;

    // Clear log and output files if present
    clearLogFiles();

    // Get the port number to use
    portNum = getNextAvailablePort(startingPortNum);
    if (portNum == -1) {
      throw new ShearRuntimeException("Could not locate a port number to use for gfserver");
    }

    // The ProcessBuilder object to run this GFserver with
    pb =
        new ProcessBuilder("gfServer", "start", "localhost", "" + portNum, twoBitFile, "-canStop", "-log=" + prefix
            + ".gfServer.log");

  }

  /**
   * Finds the next available port number to use for this GFServer, starting
   * with the specified port number and attempting up until port number 65535.
   * Returns -1 if no valid port is available.
   *
   * @param startingPortNum
   *          The first port number to check
   * @return The next available port number, or -1 if no valid port is found
   */
  private int getNextAvailablePort(int startingPortNum) {

    // Used to store results of pgrep and netstat commands, respectively
    String processCheck;
    String portCheck;

    try {

      // Get results of pgrep to check for currently-running instances of gfserver
      processCheck = RunSystemCommand.run(new ProcessBuilder("pgrep", "-l", "-f", "gfServer"));

      // Get results of netstat to check for ports being used
      portCheck = RunSystemCommand.run(new ProcessBuilder("netstat", "-an"));

      // Handles potential errors from running pgrep or netstat
    } catch (ShearRuntimeException e) {
      throw new ShearRuntimeException(
          "Unable to execute pgrep and/or netstat commands to find a valid gfserver port",
          e);
    }

    // Loop through port numbers to check
    for (int i = startingPortNum; i <= 65535; i++) {
      if (portCheck.contains("" + i) || processCheck.contains("" + i)) {
        continue;
      } else {
        return i;
      }
    }

    // Return -1 if no valid port number is found
    return -1;

  }

  @Override
  public void run() {

    try {

      // Redirect process's error stream to output
      pb.redirectErrorStream(true);

      // Start the gfserver process
      proc = pb.start();

      // Get the process output
      InputStream in = proc.getInputStream();

      // Set up output file to write output to
      FileOutputStream out = new FileOutputStream(prefix + ".gfServer.out");

      // Write output to output stream
      byte[] buf = new byte[1024];
      int len;
      while ((len = in.read(buf)) > 0) {
        out.write(buf, 0, len);
        out.flush();
      }

      // Close streams and wait for process to terminate
      in.close();
      out.close();
      proc.waitFor();

      // Catch any errors
    } catch (Exception e) {
      throw new ShearRuntimeException("Error while running GFServer", e);
    }

  }

  /**
   * Waits until the GFServer is ready and returns true. Returns false if an
   * error is encountered while waiting for the GFServer to be set up.
   *
   * @return True when GFServer is ready, false if a problem is encountered.
   * @throws InterruptedException
   *           If the thread is interrupted by another thread while it is
   *           waiting
   * @throws FileNotFoundException
   *           If the GFServer log file is not found
   */
  public boolean waitForReady() throws InterruptedException, FileNotFoundException {

    // Wait for the process to start
    while (proc == null) {
      sleep(1000);
    }

    // Whether or not the log file says that GFServer is ready
    boolean logFileReady = false;

    // The result of checking the GFServer
    String serverCheck;

    // Get result of GFServer check
    try {
      serverCheck = RunSystemCommand.run(new ProcessBuilder("gfServer", "files", "localhost", "" + portNum));
    } catch (ShearRuntimeException e) {
      throw new ShearRuntimeException("Unable to execute gfserver command to check on gfserver setup", e);
    }

    // Loop every one second until log file says GFServer is ready and GFServer check succeeds
    while (!logFileReady || !serverCheck.trim().equals(twoBitFile)) {
      sleep(1000);

      // Loop through first 10 lines of GFServer log file
      Scanner sc = new Scanner(new File(prefix + ".gfServer.log"));
      int ctr = 0;
      while (ctr < 10 && sc.hasNextLine()) {
        ctr++;
        String line = sc.nextLine();

        // Return false if socket/connection error in log file
        if (line.contains("Couldn't bind socket to") || line.contains("Error accepting the connection")) {
          sc.close();
          return false;

          // Return false and set validTwoBitFile variable if *.2bit file is invalid
        } else if (line.contains("doesn't have a valid twoBitSig") || line.contains("Unrecognized file type")) {
          validTwoBitFile = false;
          sc.close();
          return false;

          // Return true if server is ready for queries
        } else if (line.contains("Server ready for queries")) {
          logFileReady = true;
        }

      }

      // Close scanner for GFServer log file
      sc.close();

      // Get result of GFServer check
      try {
        serverCheck = RunSystemCommand.run(new ProcessBuilder("gfServer", "files", "localhost", "" + portNum));
      } catch (ShearRuntimeException e) {
        throw new ShearRuntimeException("Unable to execute gfserver command to check on gfserver setup", e);
      }

    }

    // Return true when GFServer is ready
    return true;

  }

  /**
   * Returns the *.2bit file used by this GFServer.
   *
   * @return The *.2bit file used by this GFServer
   */
  public String getTwoBitFile() {
    return twoBitFile;
  }

  /**
   * Returns the port number used by this GFServer.
   *
   * @return The port number used by this GFServer
   */
  public int getPortNum() {
    return portNum;
  }

  /**
   * Returns true if the *.2bit file used by this GFServer is valid, false
   * otherwise.
   *
   * @return True if the *.2bit file used by this GFServer is valid, false
   *         otherwise
   */
  public boolean isValidTwoBitFile() {
    return validTwoBitFile;
  }

  /**
   * Closes this GFServer by destroying the process if it is currently running.
   *
   * @throws InterruptedException
   *           If the thread is interrupted by another thread while it is
   *           waiting
   */
  public void close() throws InterruptedException {
    if (proc != null) {
      proc.destroy();
    }
  }

  /**
   * Deletes the log and output files for this GFServer.
   */
  public void clearLogFiles() {
    new File(prefix + ".gfServer.log").delete();
    new File(prefix + ".gfServer.out").delete();
  }

}
