/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.variants;

import java.io.IOException;

import edu.umn.cs.kumarbio.Genome;
import edu.umn.cs.kumarbio.shear.ShearRuntimeException;
import edu.umn.cs.kumarbio.shear.heterogeneity.Breakpoint;
import edu.umn.cs.kumarbio.shear.heterogeneity.BreakpointReads;

/**
 * Represents a tandem duplication in a genome.
 */
public class TandemDuplication extends Variant {

  // The size of this tandem duplication (i.e. how many of the original bases are duplicated)
  private long size;

  /**
   * Creates a new TandemDuplication object.
   *
   * @param loc
   *          The location of this tandem duplication (i.e. first base on the
   *          chromosome that is duplicated)
   * @param size
   *          The size of this tandem duplication (i.e. how many of the original
   *          bases are duplicated)
   */
  public TandemDuplication(Location loc, long size) {
    super(loc);

    // Parse arguments and throw errors
    if (size < 1) {
      throw new IllegalVariantException("Invalid size specified for tandem duplication (non-positive value): " + size
          + "bp");
    } else if (loc.getCoord() + size - 1 > loc.getChr().getLength()) {
      throw new IllegalVariantException("Invalid size specified for tandem duplication (too large for chromosome): "
          + loc.getChr().getName() + ":" + loc + " (" + size + "bp)");
    }

    // Initialize instance variable
    this.size = size;

    // Get the bases immediately before and after both breakpoints (up to MAX_MICROHOMOLOGY_WINDOW)
    String bp1fivePrimeBases = getFivePrimeBases(getChr(), getCoord() - 1);
    String bp1threePrimeBases = getThreePrimeBases(getChr(), getCoord() - 1);
    String bp2fivePrimeBases = getFivePrimeBases(getChr(), getEndLoc().getCoord());
    String bp2threePrimeBases = getThreePrimeBases(getChr(), getEndLoc().getCoord());

    // Find the number of bases that are identical on the 5' side of the two breakpoints
    int numFivePrimeIdenticalBases =
        getNumIdenticalBases(Genome.reverse(bp1fivePrimeBases), Genome.reverse(bp2fivePrimeBases));

    // Find the number of bases that are identical on the 3' side of the two breakpoints
    int numThreePrimeIdenticalBases = getNumIdenticalBases(bp1threePrimeBases, bp2threePrimeBases);

    // Define the range of possible breakpoint locations (location will represent the coordinate just before the break)
    long bp1fivePrimeEdge = getCoord() - 1 - numFivePrimeIdenticalBases;
    long bp1threePrimeEdge = getCoord() - 1 + numThreePrimeIdenticalBases;
    long bp2fivePrimeEdge = getEndCoord() - numFivePrimeIdenticalBases;
    long bp2threePrimeEdge = getEndCoord() + numThreePrimeIdenticalBases;

    // Validate that none of the possible breakpoint locations are negative
    if (bp1fivePrimeEdge < 0l) {
      throw new IllegalVariantException("Illegal breakpoint coordinate: " + bp1fivePrimeEdge);
    }
    if (bp1threePrimeEdge < 0l) {
      throw new IllegalVariantException("Illegal breakpoint coordinate: " + bp1threePrimeEdge);
    }
    if (bp2fivePrimeEdge < 0l) {
      throw new IllegalVariantException("Illegal breakpoint coordinate: " + bp2fivePrimeEdge);
    }
    if (bp2threePrimeEdge < 0l) {
      throw new IllegalVariantException("Illegal breakpoint coordinate: " + bp2threePrimeEdge);
    }

    // Check for any possible breakpoint locations that are before the beginning of the chromosome (i.e. 0 coordinate)
    boolean bp1fivePrimeEdgeIsBeforeStart = bp1fivePrimeEdge == 0l;
    boolean bp1threePrimeEdgeIsBeforeStart = bp1threePrimeEdge == 0l;
    boolean bp2fivePrimeEdgeIsBeforeStart = bp2fivePrimeEdge == 0l;
    boolean bp2threePrimeEdgeIsBeforeStart = bp2threePrimeEdge == 0l;

    // Set the breakpoints
    getBreakpoints().add(
        new Breakpoint(new Location(getChr(), Math.max(1l, bp1fivePrimeEdge)), new Location(getChr(), Math.max(
            1l,
            bp1threePrimeEdge)), bp1fivePrimeEdgeIsBeforeStart, bp1threePrimeEdgeIsBeforeStart));
    getBreakpoints().add(
        new Breakpoint(new Location(getChr(), Math.max(1l, bp2fivePrimeEdge)), new Location(getChr(), Math.max(
            1l,
            bp2threePrimeEdge)), bp2fivePrimeEdgeIsBeforeStart, bp2threePrimeEdgeIsBeforeStart));

  }

  /**
   * Returns the size of this tandem duplication (i.e. how many of the original
   * bases are duplicated).
   *
   * @return The size of this tandem duplication (i.e. how many of the original
   *         bases are duplicated)
   */
  public long getSize() {
    return size;
  }

  /**
   * Returns the end location of this tandem duplication (i.e. last base on the
   * chromosome that is duplicated).
   *
   * @return The end location of this tandem duplication (i.e. last base on the
   *         chromosome that is duplicated)
   */
  public Location getEndLoc() {
    return new Location(getChr(), getEndCoord());
  }

  /**
   * Returns the end coordinate of this tandem duplication (i.e. last base on
   * the chromosome that is duplicated).
   *
   * @return The end coordinate of this tandem duplication (i.e. last base on
   *         the chromosome that is duplicated)
   */
  public long getEndCoord() {
    return getCoord() + getSize() - 1;
  }

  /**
   * Returns the string of duplicated bases for this tandem duplication.
   *
   * @return The string of duplicated bases for this tandem duplication
   */
  public String getDuplicatedBases() {
    try {
      return getChr().getSequence(getCoord(), getEndCoord());
    } catch (IOException e) {
      throw new ShearRuntimeException("Error reading from the FASTA reference file", e);
    }
  }

  @Override
  public double getEstimatedReferenceDepth() {
    if (getBreakpoints().size() != 2) {
      throw new ShearRuntimeException("Exactly two breakpoints are expected for a tandem duplication");
    }
    BreakpointReads bpReads1 = getBreakpoints().get(0).getBpReads();
    BreakpointReads bpReads2 = getBreakpoints().get(1).getBpReads();
    return Math.max(
        0d,
        ((double) bpReads1.getFivePrimeSpanningReads() + (double) bpReads2.getThreePrimeSpanningReads()) / 2d
            - getEstimatedVariantDepth());
  }

  @Override
  public double getEstimatedVariantDepth() {
    if (getBreakpoints().size() != 2) {
      throw new ShearRuntimeException("Exactly two breakpoints are expected for a tandem duplication");
    }
    BreakpointReads bpReads1 = getBreakpoints().get(0).getBpReads();
    BreakpointReads bpReads2 = getBreakpoints().get(1).getBpReads();
    return (double) bpReads1.getFivePrimeClippedReads() + (double) bpReads2.getThreePrimeClippedReads();
  }

  @Override
  public TandemDuplication getCanonicalEquivalent() {
    return new TandemDuplication(new Location(getChr(), getBreakpoints().get(0).getFivePrimeCoord() + 1), getSize());
  }

  @Override
  public long getBreakpointMicrohomology() {
    long bp1microhomology = getBreakpoints().get(0).getThreePrimeCoord() - getBreakpoints().get(0).getFivePrimeCoord();
    long bp2microhomology = getBreakpoints().get(1).getThreePrimeCoord() - getBreakpoints().get(1).getFivePrimeCoord();
    if (bp1microhomology != bp2microhomology) {
      throw new ShearRuntimeException("Unexpected behavior while calculating breakpoint microhomology");
    }
    return bp1microhomology;
  }

  @Override
  public String getVariantType() {
    return "TDUP";
  }

  @Override
  public String getVariantDetails() {
    return String.format("%s:%d-%d (%d bp)", getChr().getName(), getCoord(), getEndCoord(), getSize());
  }

  @Override
  public String getSdiString() {
    return formatSdiLine(getLoc(), "-", getDuplicatedBases());
  }

  @Override
  public String getVariantIdString() {
    return super.getVariantIdString() + "\tTDUP\t" + getSize();
  }

  @Override
  public boolean equals(Object obj) {
    if (super.equals(obj)) {
      if (obj == null || !(obj instanceof TandemDuplication)) {
        return false;
      } else {
        TandemDuplication other = (TandemDuplication) obj;
        return getSize() == other.getSize();
      }
    } else {
      return false;
    }
  }

}
