/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.variants;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import edu.umn.cs.kumarbio.Chromosome;
import edu.umn.cs.kumarbio.shear.ShearRuntimeException;
import edu.umn.cs.kumarbio.shear.heterogeneity.Breakpoint;

/**
 * Represents a variant in a genome. This is an abstract class that is extended
 * for particular types of variants.
 */
public abstract class Variant implements Comparable<Variant> {

  // The maximum window size to look for microhomology on either side of a breakpoint
  public static final int  MAX_MICROHOMOLOGY_WINDOW = 100;

  // The location of this variant (i.e. first affected base in the chromosome)
  private Location         loc;

  // The heterogeneity level of this variant (-1 indicates it is not set yet)
  private double           heterogeneity;

  // The breakpoints associated with this variant
  private List<Breakpoint> breakpoints;

  // Indicates whether or not this variant is an indel (i.e. small insertion or deletion)
  private boolean          isIndel;

  /**
   * Creates a new Variant object, initializing the isIndel parameter to false,
   * the heterogeneity level to -1 (i.e. not set yet), and the breakpoints to an
   * empty list.
   *
   * @param loc
   *          The location of this variant (i.e. first affected base in the
   *          chromosome)
   */
  public Variant(Location loc) {
    this(loc, false);
  }

  /**
   * Creates a new Variant object, initializing the heterogeneity level to -1
   * (i.e. not set yet), and the breakpoints to an empty list.
   *
   * @param loc
   *          The location of this variant (i.e. first affected base in the
   *          chromosome)
   * @param isIndel
   *          Whether or not this variant is an indel (i.e. small insertion or
   *          deletion)
   */
  public Variant(Location loc, boolean isIndel) {

    // Parse arguments and throw errors
    if (loc == null) {
      throw new IllegalVariantException("Null location specified for variant");
    }

    // Initialize variables
    this.loc = loc;
    heterogeneity = -1d;
    breakpoints = new ArrayList<Breakpoint>();
    this.isIndel = isIndel;

  }

  /**
   * Returns the chromosome that this variant occurs on.
   *
   * @return The chromosome that this variant occurs on
   */
  public Chromosome getChr() {
    return getLoc().getChr();
  }

  /**
   * Returns the coordinate that this variant occurs at (i.e. first affected
   * base in the chromosome).
   *
   * @return The coordinate that this variant occurs at (i.e. first affected
   *         base in the chromosome)
   */
  public long getCoord() {
    return getLoc().getCoord();
  }

  /**
   * Returns the location of this variant (i.e. first affected base in the
   * chromosome).
   *
   * @return The location of this variant (i.e. first affected base in the
   *         chromosome)
   */
  public Location getLoc() {
    return loc;
  }

  /**
   * Returns the heterogeneity level of this variant. This will be -1 if the
   * heterogeneity level has not been set yet.
   *
   * @return The heterogeneity level of this variant
   */
  public double getHeterogeneity() {
    return heterogeneity;
  }

  /**
   * Returns the breakpoints associated with this variant. The first breakpoint
   * in the list (if there is at least one, i.e. not SNP) will be associated
   * with the location of the variant.
   *
   * @return The breakpoints associated with this variant
   */
  public List<Breakpoint> getBreakpoints() {
    return breakpoints;
  }

  /**
   * Returns true if this variant is an indel (i.e. small insertion or
   * deletion), false otherwise.
   *
   * @return True if this variant is an indel (i.e. small insertion or
   *         deletion), false otherwise
   */
  public boolean isIndel() {
    return isIndel;
  }

  /**
   * Sets the heterogeneity level of this variant.
   *
   * @param heterogeneity
   *          The heterogeneity level to set for this variant (must be between 0
   *          and 1, inclusive)
   */
  public void setHeterogeneity(double heterogeneity) {
    if (heterogeneity < 0d || heterogeneity > 1d) {
      throw new IllegalArgumentException("Invalid heterogeneity level set for variant: " + heterogeneity);
    } else {
      this.heterogeneity = heterogeneity;
    }
  }

  /**
   * Calculates and sets the heterogeneity level of this variant based on the
   * breakpoint reads information currently stored.
   */
  public void calculateHeterogeneity() {
    setHeterogeneity(Math.min(1d, getEstimatedVariantDepth()
        / (getEstimatedVariantDepth() + getEstimatedReferenceDepth())));
  }

  /**
   * Returns the estimated reference depth for this variant based on the
   * breakpoint reads information currently stored.
   *
   * @return The estimated reference depth for this variant
   */
  public abstract double getEstimatedReferenceDepth();

  /**
   * Returns the estimated variant depth for this variant based on the
   * breakpoint reads information currently stored.
   *
   * @return The estimated variant depth for this variant
   */
  public abstract double getEstimatedVariantDepth();

  /**
   * Returns the canonical equivalent of this variant. This is to account for
   * situations in which a variant (given its consensus signature) could have
   * multiple possible breakpoint due to microhomology at the breakpoints. If
   * multiple variant breakpoints would result in the same consensus signature,
   * then they should map to the same variant object here.
   *
   * @return The canonical equivalent of this variant
   */
  public abstract Variant getCanonicalEquivalent();

  /**
   * Returns the number of bases that are part of a breakpoint microhomology for
   * this variant.
   *
   * @return The number of bases that are part of a breakpoint microhomology for
   *         this variant
   */
  public abstract long getBreakpointMicrohomology();

  /**
   * Returns a string containing the type of this variant (i.e. DEL, SNP, TDUP,
   * etc.).
   *
   * @return A string containing the type of this variant
   */
  public abstract String getVariantType();

  /**
   * Returns a string containing the variant details (i.e. location, size,
   * etc.).
   *
   * @return A string containing the variant details
   */
  public abstract String getVariantDetails();

  /**
   * Returns the bases immediately to the 5' side of the specified breakpoint
   * location (up to MAX_MICROHOMOLOGY_WINDOW). The breakpoint location
   * specified should refer to the coordinate directly before the break (or 0 to
   * refer to a break that occurs before the beginning of the chromosome)
   *
   * @param chr
   *          The chromosome that the breakpoint occurs on
   * @param coord
   *          The coordinate directly before the break (or 0 to refer to a break
   *          that occurs before the beginning of the chromosome)
   * @return The bases immediately to the 5' side of the specified breakpoint
   *         location (up to MAX_MICROHOMOLOGY_WINDOW)
   */
  public static String getFivePrimeBases(Chromosome chr, long coord) {
    try {
      if (coord == 0l) {
        return "";
      } else if (coord > 0l && coord <= chr.getLength()) {
        return chr.getSequence(Math.max(1l, coord + 1 - MAX_MICROHOMOLOGY_WINDOW), coord);
      } else {
        throw new IllegalVariantException("Illegal breakpoint coordinate: " + chr.getName() + ":" + coord);
      }
    } catch (IOException e) {
      throw new ShearRuntimeException("I/O error occurred while reading from genome file");
    }
  }

  /**
   * Returns the bases immediately to the 3' side of the specified breakpoint
   * location (up to MAX_MICROHOMOLOGY_WINDOW). The breakpoint location
   * specified should refer to the coordinate directly before the break (or 0 to
   * refer to a break that occurs before the beginning of the chromosome)
   *
   * @param chr
   *          The chromosome that the breakpoint occurs on
   * @param coord
   *          The coordinate directly before the break (or 0 to refer to a break
   *          that occurs before the beginning of the chromosome)
   * @return The bases immediately to the 3' side of the specified breakpoint
   *         location (up to MAX_MICROHOMOLOGY_WINDOW)
   */
  public static String getThreePrimeBases(Chromosome chr, long coord) {
    try {
      if (coord == chr.getLength()) {
        return "";
      } else if (coord >= 0l && coord < chr.getLength()) {
        return chr.getSequence(coord + 1l, Math.min(chr.getLength(), coord + MAX_MICROHOMOLOGY_WINDOW));
      } else {
        throw new IllegalVariantException("Illegal breakpoint coordinate: " + chr.getName() + ":" + coord);
      }
    } catch (IOException e) {
      throw new ShearRuntimeException("I/O error occurred while reading from genome file");
    }
  }

  /**
   * Returns the number of identical bases (case ignored) at the beginning of
   * the two specified strings.
   *
   * @param s1
   *          The first string to compare
   * @param s2
   *          The second string to compare
   * @return The number of identical bases (case ignored) at the beginning of
   *         the two specified strings
   */
  public static int getNumIdenticalBases(String s1, String s2) {
    int numIdenticalBases = 0;
    int s1index = 0;
    int s2index = 0;
    while (s1index <= s1.length() - 1 && s2index <= s2.length() - 1
        && String.valueOf(s1.charAt(s1index)).toUpperCase().equals(String.valueOf(s2.charAt(s2index)).toUpperCase())) {
      numIdenticalBases++;
      s1index++;
      s2index++;
    }
    return numIdenticalBases;
  }

  /**
   * Returns a string containing a properly-formatted SDI line for a particular
   * variant specified by a location, string of reference bases, and string of
   * consensus (i.e. variant) bases. This method will format the information
   * into SDI format and calculate the variant size difference, which is the
   * third field in an SDI line. This also validates the reference bases to
   * ensure that they match the bases present at the specified location.
   *
   * @param loc
   *          The location of this variant
   * @param reference
   *          The reference bases at this location (can be "-" to specify no
   *          changed reference bases)
   * @param consensus
   *          The consensus/variant bases that are replacing the reference bases
   *          at this location (can be "-" to specify a deletion of reference
   *          bases)
   * @return A properly-formatted SDI line for the specified variant information
   */
  public static String formatSdiLine(Location loc, String reference, String consensus) {

    // Validate reference bases
    try {
      if (!reference.equals("-")) {
        if (!loc.getChr().getSequence(loc.getCoord(), loc.getCoord() + reference.length() - 1).equals(reference)) {
          throw new ShearRuntimeException("Illegal SDI line specified (reference bases to not match location)");
        }
      }
    } catch (IOException e) {
      throw new ShearRuntimeException("I/O error occurred while validating SDI line");
    }

    // StringBuffer to hold SDI line
    StringBuffer line = new StringBuffer();

    // Chromosome name and coordinate
    line.append(loc.getChr().getName());
    line.append("\t");
    line.append(loc.getCoord());
    line.append("\t");

    // Calculate length of variant (i.e. difference between consensus and reference sequences)
    if (reference.equals("-")) {
      if (consensus.equals("-")) {
        line.append(0);
      } else {
        line.append(consensus.length());
      }
    } else if (consensus.equals("-")) {
      if (reference.equals("-")) {
        line.append(0);
      } else {
        line.append("-");
        line.append(reference.length());
      }
    } else {
      line.append(consensus.length() - reference.length());
    }
    line.append("\t");

    // Reference and consensus bases and end of SDI string
    line.append(reference);
    line.append("\t");
    line.append(consensus);
    line.append("\t*\t*");

    // Return SDI line
    return line.toString();

  }

  /**
   * Returns a string containing the SDI format of this variant.
   *
   * @return A string containing the SDI format of this variant
   */
  public abstract String getSdiString();

  /**
   * This returns a string that contains all of the information about this
   * variant that would make it unique (used in compareTo method).
   *
   * @return The variant ID string for this variant
   */
  public String getVariantIdString() {
    StringBuffer variantIdString = new StringBuffer();
    variantIdString.append(getLoc() + "\t" + getHeterogeneity() + "\t");
    for (Breakpoint bp : getBreakpoints()) {
      variantIdString.append("[" + bp + "]" + "\t");
    }
    variantIdString.append(isIndel());
    return variantIdString.toString();
  }

  @Override
  public int compareTo(Variant other) {

    // Return 0 if variants are equal
    if (equals(other)) {
      return 0;
    }

    // Return whichever has first location
    if (getLoc().compareTo(other.getLoc()) < 0) {
      return -1;
    } else if (getLoc().compareTo(other.getLoc()) > 0) {
      return +1;
    }

    // If exact same location, then return based on variant ID string
    int comparison = getVariantIdString().compareTo(other.getVariantIdString());
    if (comparison == 0) {
      throw new ShearRuntimeException(
          "Unexpected behavior while comparing order of Variant objects (variant ID string comparison returned 0)");
    }
    return comparison;

  }

  @Override
  public boolean equals(Object obj) {
    if (obj == null || !(obj instanceof Variant)) {
      return false;
    } else {
      Variant other = (Variant) obj;
      if (getBreakpoints().size() == other.getBreakpoints().size()) {
        for (int i = 0; i < getBreakpoints().size(); i++) {
          if (!getBreakpoints().get(i).equals(other.getBreakpoints().get(i))) {
            return false;
          }
        }
        return getLoc().equals(other.getLoc()) && getHeterogeneity() == other.getHeterogeneity()
            && isIndel() == other.isIndel();
      } else {
        return false;
      }
    }
  }

  @Override
  public String toString() {
    if (getBreakpointMicrohomology() > 0l) {
      return String.format(
          "%s %s (%d bp microhomology)",
          getVariantType(),
          getVariantDetails(),
          getBreakpointMicrohomology());
    } else {
      return String.format("%s %s", getVariantType(), getVariantDetails());
    }
  }

}
