/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.variants;

import java.io.IOException;

import edu.umn.cs.kumarbio.shear.ShearRuntimeException;

/**
 * Represents a SNP in a genome.
 */
public class SNP extends Variant {

  // The reference base that is being substituted for
  private char refBase;

  // The new base that is being substituted in
  private char varBase;

  /**
   * Creates a new SNP object.
   *
   * @param loc
   *          The location of this SNP (i.e. the base that is substituted)
   * @param varBase
   *          The new base that is being substituted in
   */
  public SNP(Location loc, char varBase) {
    super(loc);

    // Parse arguments and throw errors
    if (varBase != 'A' && varBase != 'C' && varBase != 'G' && varBase != 'T') {
      throw new IllegalVariantException("Invalid variant based specified for SNP: " + varBase);
    }

    // Initialize instance variables
    try {
      refBase = loc.getChr().getSequence(loc.getCoord(), loc.getCoord()).charAt(0);
    } catch (IOException e) {
      throw new ShearRuntimeException("I/O error occurred while reading from genome file");
    }
    this.varBase = varBase;

  }

  /**
   * Returns the reference base that is being substituted for.
   *
   * @return The reference base that is being substituted for
   */
  public char getRefBase() {
    return refBase;
  }

  /**
   * Returns the new base that is being substituted in
   *
   * @return The new base that is being substituted in
   */
  public char getVarBase() {
    return varBase;
  }

  @Override
  public double getEstimatedReferenceDepth() {
    throw new ShearRuntimeException("Cannot estimate reference depth for a SNP object");
  }

  @Override
  public double getEstimatedVariantDepth() {
    throw new ShearRuntimeException("Cannot estimate variant depth for a SNP object");
  }

  @Override
  public SNP getCanonicalEquivalent() {
    return new SNP(getLoc(), getVarBase());
  }

  @Override
  public long getBreakpointMicrohomology() {
    return 0;
  }

  @Override
  public String getVariantType() {
    return "SNP";
  }

  @Override
  public String getVariantDetails() {
    return String.format("%s:%d (%s>%s)", getChr().getName(), getCoord(), getRefBase(), getVarBase());
  }

  @Override
  public String getSdiString() {
    return formatSdiLine(getLoc(), String.valueOf(getRefBase()), String.valueOf(getVarBase()));
  }

  @Override
  public String getVariantIdString() {
    return super.getVariantIdString() + "\tSNP\t" + getRefBase() + "\t" + getVarBase();
  }

  @Override
  public boolean equals(Object obj) {
    if (super.equals(obj)) {
      if (obj == null || !(obj instanceof SNP)) {
        return false;
      } else {
        SNP other = (SNP) obj;
        return getRefBase() == other.getRefBase() && getVarBase() == other.getVarBase();
      }
    } else {
      return false;
    }
  }

}
